#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import sys
from os import listdir
from os.path import isfile, join
from typing import Optional

import numpy as np
import tifffile.tifffile as tif

from tardis_em.utils.errors import TardisError


class StitchImages:
    """
    MAIN MODULE TO STITCH IMAGE FROM IMAGE PATCHES

    Class object to stitch cut date into one big image. Object recognize images
    with naming 0_1_1_1_25_pf where:
    - 0 indicate id of image
    - 1 indicate xyz position
    - 25 indicate stride value for overlap
    - pf indicate optional prefix in file name
    """

    def __init__(self):
        self.idx = 0  # Variable storing number of stitched images
        self.nx, self.ny, self.nz = (
            0,
            0,
            0,
        )  # Variable used to store xyz image dimension
        self.x, self.y, self.z = 0, 0, 0  # Variable to store number of patches in xyz
        self.stride = 0  # Variable to store step size

    def _find_xyz(self, file_list: list, idx: int):
        """
        Find index from for stitching image patches into one file.

        Args:
            file_list (list): List of files.
            idx: Find file index number.

        Returns:
            Update global class values.
        """
        self.z = (
            max(
                list(
                    map(
                        int,
                        [
                            str.split(f[:-4], "_")[1]
                            for f in file_list
                            if f.startswith(f"{idx}")
                        ],
                    )
                )
            )
            + 1
        )
        self.y = (
            max(
                list(
                    map(
                        int,
                        [
                            str.split(f[:-4], "_")[2]
                            for f in file_list
                            if f.startswith(f"{idx}")
                        ],
                    )
                )
            )
            + 1
        )
        self.x = (
            max(
                list(
                    map(
                        int,
                        [
                            str.split(f[:-4], "_")[3]
                            for f in file_list
                            if f.startswith(f"{idx}")
                        ],
                    )
                )
            )
            + 1
        )
        self.stride = max(
            list(
                map(
                    int,
                    [
                        str.split(f[:-4], "_")[4]
                        for f in file_list
                        if f.startswith(f"{idx}")
                    ],
                )
            )
        )

    def _calculate_dim(self, image: np.ndarray):
        """
        Find and update image patch size from array.

        Args:
            image (np.ndarray): Image array.

        Returns:
            Update global class values.
        """
        if image.ndim == 3:
            self.nz, self.ny, self.nx = image.shape
        else:
            self.ny, self.nx = image.shape
            self.nz = 0

    def __call__(
        self,
        image_dir: str,
        mask: bool,
        output: Optional[str] = None,
        prefix="",
        dtype=np.uint8,
    ) -> np.ndarray:
        """
        STITCH IMAGE FROM IMAGE PATCHES

        Args:
            image_dir (str): Directory where all images are stored.
            mask (np.ndarray): If True treat image as binary mask and sum-up overlay
                zones, else do replacement.
            output (str, Optional): Optional, output directory.
            dtype (np.dtype): Numpy dtype for output
            prefix (str): Prefix name if available.

        Returns:
            np.ndarray: If indicated output, image is saved in output directory
            else stitch images is return as array.
        """
        """Extract information about images in dir_path"""
        file_list = [f for f in listdir(image_dir) if isfile(join(image_dir, f))]
        file_list = [f for f in file_list if f.endswith(".tif")]

        # Number of images to stitch
        self.idx = (
            max(list(map(int, [str.split(f[:-4], "_")[0] for f in file_list]))) + 1
        )

        for idx in range(self.idx):
            self._find_xyz(file_list, idx)
            self._calculate_dim(
                tif.imread(join(image_dir, f"{idx}_0_0_0_{self.stride}{prefix}.tif"))
            )

            x_dim = self.nx + ((self.nx - self.stride) * (self.x - 1))
            y_dim = self.ny + ((self.ny - self.stride) * (self.y - 1))
            if self.nz == 0:
                z_dim = 0
                stitched_image = np.zeros((y_dim, x_dim), dtype=dtype)
            else:
                z_dim = self.nz + ((self.nz - self.stride) * (self.z - 1))
                stitched_image = np.zeros((z_dim, y_dim, x_dim), dtype=dtype)

            z_start, z_stop = 0 - (self.nz - self.stride), 0

            if self.z == 0:
                self.z = 1

            batch_iter_z = range(self.z)

            batch_iter_y = range(self.y)

            for i in batch_iter_z:
                z_start = z_start + self.nz - self.stride
                z_stop = z_start + self.nz
                y_start, y_stop = 0 - (self.ny - self.stride), 0

                for j in batch_iter_y:
                    y_start = y_start + self.ny - self.stride
                    y_stop = y_start + self.ny
                    x_start, x_stop = 0 - (self.nx - self.stride), 0

                    for k in range(self.x):
                        x_start = x_start + self.nx - self.stride
                        x_stop = x_start + self.nx

                        img_name = str(
                            join(
                                image_dir,
                                f"{idx}_{i}_{j}_{k}_{self.stride}{prefix}.tif",
                            )
                        )

                        img = tif.imread(img_name)

                        if self.nz == 0:
                            if img.shape != (self.ny, self.nx):
                                TardisError(
                                    id_="",
                                    py="tardis_em/cnn/data_processing.md/stitch.py",
                                    desc=f"Stitch image size does not match. {img.shape} "
                                    f"doesn't match ({self.ny}, {self.nx})",
                                )
                                sys.exit()
                        else:
                            if img.shape != (self.nz, self.ny, self.nx):
                                TardisError(
                                    id_="",
                                    py="tardis_em/cnn/data_processing.md/stitch.py",
                                    desc=f"Stitch image size does not match. {img.shape} "
                                    f"doesn't match ({self.nz}, {self.ny}, {self.nx})",
                                )
                                sys.exit()

                        if mask and self.nz == 0:
                            stitched_image[y_start:y_stop, x_start:x_stop] += img
                        elif mask and self.nz > 0:
                            stitched_image[
                                z_start:z_stop, y_start:y_stop, x_start:x_stop
                            ] += img
                        elif not mask and self.nz == 0:
                            stitched_image[y_start:y_stop, x_start:x_stop] += img
                        else:
                            stitched_image[
                                z_start:z_stop, y_start:y_stop, x_start:x_stop
                            ] += img

            # Reduce overlapping areas by averaging
            if mask:
                stitched_image = np.where(stitched_image > 0, 1, 0).astype(np.uint8)
            else:
                if self.nz == 0:
                    grid_y, grid_x = generate_grid(
                        stitched_image.shape,
                        [self.ny, self.nx],
                        [self.y, self.x],
                        self.stride,
                    )
                    if len(grid_y) > 0:
                        stitched_image[grid_y[:,], :] = (
                            stitched_image[grid_y[:,], :] / 2
                        )
                    if len(grid_x) > 0:
                        stitched_image[:, grid_x[:,]] = (
                            stitched_image[:, grid_x[:,]] / 2
                        )
                else:
                    grid_y, grid_x, grid_z = generate_grid(
                        stitched_image.shape,
                        [self.nz, self.ny, self.nx],
                        [self.z, self.y, self.x],
                        self.stride,
                    )

                    if len(grid_y) > 0:
                        stitched_image[grid_y[:, 0], grid_y[:, 1], :] = (
                            stitched_image[grid_y[:, 0], grid_y[:, 1], :] / 2
                        )
                    if len(grid_x) > 0:
                        stitched_image[grid_x[:, 0], :, grid_x[:, 1]] = (
                            stitched_image[grid_x[:, 0], :, grid_x[:, 1]] / 2
                        )
                    if len(grid_z) > 0:
                        stitched_image[grid_z[:,], ...] = (
                            stitched_image[grid_z[:,], ...] / 2
                        )

            if output is None:
                return np.array(stitched_image, dtype=dtype)
            else:
                tif.imwrite(
                    join(output, f"Stitched_Image_idx_{idx}.tif"),
                    np.array(stitched_image, dtype=dtype),
                )


def generate_grid(image_size: tuple, patch_size: list, grid_size: list, stride: int):
    """
    Generates grid coordinates for either 2D or 3D images.

    Args:
        image_size (list): The dimensions of the image.
            For 3D, it should be [z, y, x], and for 2D, [y, x].
        patch_size (list): The dimensions of each patch.
            For 3D, [nz, ny, nx], and for 2D, [ny, nx].
        grid_size (list): The grid size.
            For 3D, [gz, gy, gx], and for 2D, [gy, gx].
        stride (int): The stride for grid generation.

    Returns:
        Tuple[list]: A tuple of numpy arrays with coordinates of the grid.
    """
    # Determine if the image is 2D based on the length of image_size
    D2 = len(image_size) == 2

    # Unpack the dimensions based on whether the image is 2D or 3D
    if D2:
        z_dim, nz, gz = 0, 0, 0
        y_dim, x_dim = image_size
        ny, nx = patch_size
        gy, gx = grid_size
    else:
        z_dim, y_dim, x_dim = image_size
        nz, ny, nx = patch_size
        gz, gy, gx = grid_size

    # Generate y and x coordinates
    y_coords = [
        np.arange((ny * g - stride * g), ny * g - (stride * (g - 1)))
        for g in range(1, gy)
    ]
    if len(y_coords) > 0:
        y_coords = np.concatenate(
            [
                np.arange((ny * g - stride * g), ny * g - (stride * (g - 1)))
                for g in range(1, gy)
            ]
        )

    x_coords = [
        np.arange((nx * g - stride * g), nx * g - (stride * (g - 1)))
        for g in range(1, gx)
    ]
    if len(x_coords) > 0:
        x_coords = np.concatenate(
            [
                np.arange((nx * g - stride * g), nx * g - (stride * (g - 1)))
                for g in range(1, gx)
            ]
        )

    # Generate z coordinates and meshgrids if the image is 3D
    if not D2:
        z_coords = np.arange(z_dim)
        zz, yy = np.meshgrid(z_coords, y_coords, indexing="ij")
        coordinates_yy = np.vstack([zz.ravel(), yy.ravel()]).T

        zz, xx = np.meshgrid(z_coords, x_coords, indexing="ij")
        coordinates_xx = np.vstack([zz.ravel(), xx.ravel()]).T

        z_coords = [
            np.arange((nz * g - stride * g), nz * g - (stride * (g - 1)))
            for g in range(1, gz)
        ]
        if len(z_coords) > 0:
            z_coords = np.concatenate(z_coords)

        return coordinates_yy, coordinates_xx, z_coords
    else:
        return y_coords, x_coords



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from os import getcwd, listdir, mkdir
from os.path import isdir, join
from shutil import rmtree
from typing import List, Tuple, Union

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset

from tardis_em.dist_pytorch.datasets.augmentation import preprocess_data
from tardis_em.dist_pytorch.datasets.patches import PatchDataSet
from tardis_em.utils.errors import TardisError
from tardis_em.utils.load_data import (
    load_ply_partnet,
    load_ply_scannet,
    load_s3dis_scene,
)
from tardis_em.dist_pytorch.utils.utils import (
    RandomDownSampling,
    VoxelDownSampling,
)
from tardis_em.dist_pytorch.utils.utils import pc_median_dist
from tardis_em.dist_pytorch.utils.build_point_cloud import create_simulated_dataset
from tardis_em.utils.spline_metric import sort_segment


class BasicDataset(Dataset):
    """
    BASIC CLASS FOR STANDARD DATASET CONSTRUCTION

    Args:
        coord_dir (str): Dataset directory.
        coord_format (tuple, str): A tuple of allowed coord formats.
        patch_if (int): Max number of points per patch.
        train (bool): If True, compute as for training dataset, else test.
    """

    def __init__(
        self,
        coord_dir=None,
        coord_format=".csv",
        patch_if=500,
        downscale=None,
        rgb=False,
        benchmark=False,
        train=True,
    ):
        # Coord setting
        self.coord_dir = coord_dir
        self.coord_format = coord_format
        self.rgb = rgb

        self.downscale = downscale

        self.train = train
        self.benchmark = benchmark

        # Setup environment
        self.cwd = getcwd()
        self.temp = ""

        if self.train:
            if isdir(join(self.cwd, "temp_train")):
                rmtree(join(self.cwd, "temp_train"))
            mkdir(join(self.cwd, "temp_train"))
        else:
            if isdir(join(self.cwd, "temp_test")):
                rmtree(join(self.cwd, "temp_test"))
            mkdir(join(self.cwd, "temp_test"))

        # List of detected coordinates files
        if isinstance(coord_dir, str):
            self.ids = [f for f in listdir(coord_dir) if f.endswith(self.coord_format)]

            # Save patch size value for speed-up
            self.patch_size = np.zeros((len(self.ids), 1))
            self.VD = None
        else:
            self.ids = []

        # Patch build setting
        self.max_point_in_patch = patch_if

        self.l_ = len(self.ids)

    def __len__(self):
        if self.l_ > 0:
            return self.l_
        else:
            return 1

    def save_temp(self, i: int, **kwargs):
        """
        General class function to save temp data.

        Args:
            i (int): Temp index value.
            kwargs (np.ndarray): Dictionary of all arrays to save.
        """

        for key, values in kwargs.items():
            np.save(
                join(self.cwd, self.temp, f"{key}_{i}.npy"),
                np.asarray(values, dtype=object),
            )

    def load_temp(self, i: int, **kwargs) -> List[np.ndarray]:
        """
        General class function to load temp data

        Args:
            i (int): Temp index value.
            kwargs (bool): Dictionary of all arrays to load.

        Returns:
            List (list[np.ndarray]): List of kwargs arrays as a tensor arrays.
        """
        if len(kwargs.items()) == 1:
            return np.load(
                join(self.cwd, self.temp, f"{list(kwargs.items())[0][0]}_{i}.npy"),
                allow_pickle=True,
            )
        return [
            np.load(join(self.cwd, self.temp, f"{key}_{i}.npy"), allow_pickle=True)
            for key, _ in kwargs.items()
        ]

    @staticmethod
    def list_to_tensor(**kwargs) -> List[List[np.ndarray]]:
        """
        Static class function to transform a list of numpy arrays to a list of
        tensor arrays.

        Args:
            kwargs (np.ndarray): Dictionary of all files to transform into a tensor.

        Returns:
            List (list[torch.Tensor]): List of kwargs arrays as a tensor array.
        """
        return [
            [torch.Tensor(df.astype(np.float32)).type(torch.float32) for df in value]
            for _, value in kwargs.items()
        ]

    def __getitem__(self, i: int):
        pass


class FilamentSimulateDataset(BasicDataset):
    """
    SIMULATED FILAMENT-TYPE DATASET CONSTRUCTION

    Returns:
        Tuple (list[np.ndarray]):
        coords_idx: Numpy or Tensor list of coordinates (N, (2, 3)).

        df_idx: Normalize zero-out output for standardized dummy.

        graph_idx: Numpy or Tensor list of 2D GT graphs.

        output_idx: Numpy or Tensor list (N, 1) of output index value.

        df_idx: Normalize zero-out output for standardized dummy.
    """

    def __init__(self, type_: str, sample_count=50, **kwargs):
        super(FilamentSimulateDataset, self).__init__(**kwargs)

        self.sample_count = sample_count
        self.type = type_

        self.VD = PatchDataSet(
            max_number_of_points=self.max_point_in_patch,
            overlap=0.15,
            drop_rate=0.1,
            graph=True,
            tensor=True,
        )

    def __len__(self):
        return self.sample_count

    def __getitem__(self, i: int) -> Tuple[list, list, list, list, list]:
        """Get list of all coordinates and image patches"""
        if self.train:
            self.temp = "temp_train"
        else:
            self.temp = "temp_test"

        # Simulate filament dataset
        if self.type in ["mix3d", "membranes"]:
            coord_file = create_simulated_dataset(
                size=list(np.random.randint((50, 512, 512), (60, 640, 640))),
                sim_type=self.type,
            )
            mesh = 8
        else:
            coord_file = create_simulated_dataset(
                size=list(np.random.randint((1, 512, 512), (250, 4096, 4096))),
                sim_type=self.type,
            )
            mesh = 2
        # Pre-process coord and image data also, if exist remove duplicates
        coord, _ = preprocess_data(coord=coord_file)

        # Optional Down-sampling of simulated dataset
        if self.downscale is not None:
            scale = self.downscale.split("_")

            if scale[1] == "random":
                if scale[0] == "v":
                    scale[1] = np.random.randint(1, 15)
                else:
                    scale[1] = np.random.randint(25, 100) / 100
            else:
                scale[1] = float(scale[1])

            if scale[0] == "v":
                down_scale = VoxelDownSampling(voxel=scale[1], labels=True, KNN=True)
            else:
                down_scale = RandomDownSampling(
                    threshold=scale[1], labels=True, KNN=True
                )
            coord = down_scale(coord)

            coord = coord[coord[:, 0].argsort()]
            df_coord = []
            for i in np.unique(coord[:, 0]):
                id_ = i
                idx = np.where(coord[:, 0] == id_)[0]

                if len(idx) > 3:
                    df_coord.append(
                        np.hstack(
                            (
                                np.repeat(id_, len(idx)).reshape(-1, 1),
                                sort_segment(coord[idx, 1:]),
                            )
                        )
                    )
            coord = np.concatenate(df_coord)

        # Normalize distance
        # coord[:, 1:] = coord[:, 1:] / pc_median_dist(coord[:, 1:], True)

        # Patch dataset
        if self.train:
            coords_idx, df_idx, graph_idx, output_idx, _ = self.VD.patched_dataset(
                coord=coord, mesh=mesh, random=True
            )
        else:
            coords_idx, df_idx, graph_idx, output_idx, _ = self.VD.patched_dataset(
                coord=coord, mesh=mesh
            )

        # Output edge_f,   node_f, graph,     node_idx,   node_class
        return coords_idx, df_idx, graph_idx, output_idx, df_idx


class FilamentDataset(BasicDataset):
    """
    FILAMENT-TYPE DATASET CONSTRUCTION

    Returns:
        Tuple (list[np.ndarray]):
        coords_idx: Numpy or Tensor list of coordinates (N, (2, 3)).

        df_idx: Normalize zero-out output for standardized dummy.

        graph_idx: Numpy or Tensor list of 2D GT graphs.

        output_idx: Numpy or Tensor list (N, 1) of output index value.

        df_idx: Normalize zero-out output for standardized dummy.
    """

    def __init__(self, **kwargs):
        super(FilamentDataset, self).__init__(**kwargs)
        self.VD = PatchDataSet(
            max_number_of_points=self.max_point_in_patch,
            overlap=0.1,
            drop_rate=0.1,
            graph=True,
            tensor=False,
        )

    def __getitem__(self, i: int) -> Tuple[list, list, list, list, list]:
        """Get list of all coordinates and image patches"""
        idx = self.ids[i]

        if self.train:
            self.temp = "temp_train"
        else:
            self.temp = "temp_test"

        # Define what coordinate format are available
        coord_file = join(self.coord_dir, str(idx))

        if self.patch_size[i, 0] == 0:
            # Pre-process coord and image data also, if exist remove duplicates
            coord, _ = preprocess_data(coord=coord_file)

            try:
                px = float(str(idx).split("_")[0])
            except ValueError:
                px = 1

            # Normalize point cloud to pixel size
            if self.downscale is None:
                scale = 1
            else:
                scale = self.downscale
            coord[:, 1:] = coord[:, 1:] / px  # Normalize for pixel size
            coord[:, 1:] = coord[:, 1:] / scale  # in nm know distance between points

            coords_idx, df_idx, graph_idx, output_idx, _ = self.VD.patched_dataset(
                coord=coord, mesh=2
            )

            # save data for faster access later
            if not self.benchmark:
                self.save_temp(
                    i=i, coords=coords_idx, graph=graph_idx, out=output_idx, df=df_idx
                )

                # Store initial patch size for each data to speed up computation
                self.patch_size[i, 0] = 1
        else:
            # Load pre-process data
            coords_idx, graph_idx, output_idx, df_idx = self.load_temp(
                i, coords=True, graph=True, out=True, df=True
            )

        coords_idx, graph_idx, output_idx, df_idx = self.list_to_tensor(
            coord=coords_idx, graph=graph_idx, output=output_idx, df=df_idx
        )

        # Output edge_f,   node_f, graph,     node_idx,   node_class
        return coords_idx, df_idx, graph_idx, output_idx, df_idx


class PartnetDataset(BasicDataset):
    """
    PARTNET TYPE DATASET CONSTRUCTION

    Returns:
        Tuple (list[np.ndarray]):
        coords_idx: Numpy or Tensor list of coordinates (N, (2, 3)).

        df_idx: Normalize zero-out output for standardized dummy.

        graph_idx: Numpy or Tensor list of 2D GT graphs.

        output_idx: Numpy or Tensor list (N, 1) of output index value.

        df_idx: Normalize zero-out output for standardized dummy.
    """

    def __init__(self, **kwargs):
        super(PartnetDataset, self).__init__(**kwargs)
        self.VD = PatchDataSet(
            max_number_of_points=self.max_point_in_patch,
            overlap=0.1,
            drop_rate=0.1,
            graph=True,
            tensor=False,
        )

    def __getitem__(self, i: int) -> Tuple[list, list, list, list, list]:
        """Get list of all coordinates and image patches"""
        idx = self.ids[i]

        if self.train:
            self.temp = "temp_train"
        else:
            self.temp = "temp_test"

        # Define what coordinate format are available
        coord_file = join(self.coord_dir, str(idx))

        if self.patch_size[i, 0] == 0:
            if self.downscale is None:
                scale = 0.05
            else:
                scale = self.downscale

            # Pre-process coord and image data also, if exist remove duplicates
            coord = load_ply_partnet(coord_file, downscaling=scale)
            coord[:, 1:] = coord[:, 1:] / scale

            coords_idx, df_idx, graph_idx, output_idx, _ = self.VD.patched_dataset(
                coord=coord, mesh=6
            )
            # save data for faster access later
            if not self.benchmark:
                self.save_temp(
                    i=i, coords=coords_idx, graph=graph_idx, out=output_idx, df=df_idx
                )

                # Store initial patch size for each data to speed up computation
                self.patch_size[i, 0] = 1
        else:
            # Load pre-process data
            coords_idx, graph_idx, output_idx, df_idx = self.load_temp(
                i=i, coords=True, graph=True, out=True, df=True
            )

        coords_idx, graph_idx, output_idx, df_idx = self.list_to_tensor(
            coord=coords_idx, graph=graph_idx, output=output_idx, df=df_idx
        )

        # Output edge_f,   node_f, graph,     node_idx,   node_class
        return coords_idx, df_idx, graph_idx, output_idx, df_idx


class ScannetDataset(BasicDataset):
    """
    SCANNET V2 TYPE DATASET CONSTRUCTION

    Returns:
        Tuple (list[np.ndarray]):
        coords_idx: Numpy or Tensor list of coordinates (N, (2, 3)).

        df_idx: Normalize zero-out output for standardized dummy.

        graph_idx: Numpy or Tensor list of 2D GT graphs.

        output_idx: Numpy or Tensor list (N, 1) of output index value.

        df_idx: Normalize zero-out output for standardized dummy.
    """

    def __init__(self, **kwargs):
        super(ScannetDataset, self).__init__(**kwargs)
        self.VD = PatchDataSet(
            max_number_of_points=self.max_point_in_patch,
            overlap=0.1,
            drop_rate=0.1,
            graph=True,
            tensor=False,
        )

    def __getitem__(self, i: int) -> Tuple[list, list, list, list, list]:
        """Get list of all coordinates and image patches"""
        idx = self.ids[i]

        if self.train:
            self.temp = "temp_train"
        else:
            self.temp = "temp_test"

        # Define what coordinate format are available
        coord_file = join(self.coord_dir, str(idx))

        if self.patch_size[i, 0] == 0:
            if self.downscale is None:
                scale = 0.05
            else:
                scale = self.downscale

            # Pre-process coord and image data also, if exist remove duplicates
            coord = load_ply_scannet(coord_file, downscaling=scale)
            coord[:, 1:] = coord[:, 1:] / scale

            (
                coords_idx,
                df_idx,
                graph_idx,
                output_idx,
                cls_idx,
            ) = self.VD.patched_dataset(coord=coord, mesh=6)

            if not self.benchmark:
                # save data for faster access later
                self.save_temp(
                    i=i,
                    coords=coords_idx,
                    graph=graph_idx,
                    out=output_idx,
                    df=df_idx,
                    cls=cls_idx,
                )

                # Store initial patch size for each data to speed up computation
                self.patch_size[i, 0] = 1
        else:
            # Load pre-process data
            coords_idx, graph_idx, output_idx, df_idx, cls_idx = self.load_temp(
                i, coords=True, graph=True, out=True, df=True, cls=True
            )

        coords_idx, graph_idx, output_idx, df_idx, cls_idx = self.list_to_tensor(
            coord=coords_idx, graph=graph_idx, output=output_idx, df=df_idx, cls=cls_idx
        )

        # Output edge_f,   node_f, graph,     node_idx,   node_class
        return coords_idx, df_idx, graph_idx, output_idx, cls_idx


class ScannetColorDataset(BasicDataset):
    """
    SCANNET V2 + COLORS TYPE DATASET CONSTRUCTION

    Returns:
        Tuple (list[np.ndarray]):
        coords_idx: Numpy or Tensor list of coordinates (N, (2, 3)).

        rgb_idx: Numpy or Tensor list of RGB values (N, 3).

        graph_idx: Numpy or Tensor list of 2D GT graphs.

        output_idx: Numpy or Tensor list (N, 1) of output index value.

        df_idx: Normalize zero-out output for standardized dummy.
    """

    def __init__(self, **kwargs):
        super(ScannetColorDataset, self).__init__(**kwargs)
        self.color_dir = join(self.coord_dir, "../../", "color")
        self.VD = PatchDataSet(
            max_number_of_points=self.max_point_in_patch,
            overlap=0.1,
            drop_rate=0.1,
            graph=True,
            tensor=False,
        )

    def __getitem__(self, i: int) -> Tuple[list, list, list, list, list]:
        # Check if color folder exist
        if not isdir(self.color_dir):
            TardisError(
                "12",
                "tardis_em/dist_pytorch/datasets/dataloader.py",
                f"Given dir: {self.color_dir} is not a directory!",
            )

        """ Get list of all coordinates and image patches """
        idx = self.ids[i]

        if self.train:
            self.temp = "temp_train"
        else:
            self.temp = "temp_test"

        # Define what coordinate format are available
        coord_file = join(self.coord_dir, str(idx))

        if self.patch_size[i, 0] == 0:
            if self.downscale is None:
                scale = 0.05
            else:
                scale = self.downscale

            # Pre-process coord and image data also, if exist remove duplicates
            coord, rgb = load_ply_scannet(
                coord_file,
                downscaling=scale,
                color=join(self.color_dir, f"{idx[:-11]}.ply"),
            )
            coord[:, 1:] = coord[:, 1:] / scale
            classes = coord[:, 0]

            (
                coords_idx,
                rgb_idx,
                graph_idx,
                output_idx,
                cls_idx,
            ) = self.VD.patched_dataset(coord=coord, label_cls=classes, rgb=rgb, mesh=6)

            if not self.benchmark:
                # save data for faster access later
                self.save_temp(
                    i=i,
                    coords=coords_idx,
                    graph=graph_idx,
                    out=output_idx,
                    rgb=rgb_idx,
                    cls=cls_idx,
                )

                # Store initial patch size for each data to speed up computation
                self.patch_size[i, 0] = 1
        else:
            # Load pre-process data
            coords_idx, graph_idx, output_idx, rgb_idx, cls_idx = self.load_temp(
                i, coords=True, graph=True, out=True, rgb=True, cls=True
            )

        coords_idx, graph_idx, output_idx, rgb_idx, cls_idx = self.list_to_tensor(
            coord=coords_idx,
            graph=graph_idx,
            output=output_idx,
            rgb=rgb_idx,
            cls=cls_idx,
        )

        # Output edge_f,   node_f,  graph,     node_idx,   node_class
        return coords_idx, rgb_idx, graph_idx, output_idx, cls_idx


class Stanford3DDataset(BasicDataset):
    """
    S3DIS TYPE DATASET CONSTRUCTION

    Returns:
        Tuple (list[np.ndarray]):
        coords_idx: Numpy or Tensor list of coordinates (N, (2, 3)).

        rgb_idx: Numpy or Tensor list of RGB values (N, 3).

        graph_idx: Numpy or Tensor list of 2D GT graphs.

        output_idx: Numpy or Tensor list (N, 1) of output index value.

        df_idx: Normalize zero-out output for standardized dummy.
    """

    def __init__(self, **kwargs):
        super(Stanford3DDataset, self).__init__(**kwargs)

        # Modified self.ids to recognize folder for .txt
        area_list = [
            d for d in listdir(self.coord_dir) if isdir(join(self.coord_dir, d))
        ]
        area_list = [f for f in area_list if not f.startswith(".")]

        self.ids = []
        for i in area_list:
            area = join(self.coord_dir, i)
            room_list = [d for d in listdir(area) if isdir(join(area, d))]
            room_list = [join(i, f) for f in room_list if not f.startswith(".")]

            self.ids.append(room_list)
        self.ids = [item for sublist in self.ids if sublist for item in sublist]

        # Save patch size value for speed-up
        self.patch_size = np.zeros((len(self.ids), 1))

        self.VD = PatchDataSet(
            max_number_of_points=self.max_point_in_patch,
            overlap=0,
            drop_rate=1,
            graph=True,
            tensor=False,
        )

    def __getitem__(self, i: int):
        """Get list of all coordinates and image patches"""
        idx = self.ids[i]

        if self.train:
            self.temp = "temp_train"
        else:
            self.temp = "temp_test"

        # Iterate throw every folder and generate S3DIS scene
        coord_file = join(self.coord_dir, idx, "Annotations")

        if self.patch_size[i, 0] == 0:
            print(f"Loading: {idx}")
            # start = time.time()
            # Pre-process coord and image data also, if exist remove duplicates
            if self.rgb:
                if self.downscale is not None:
                    scale = self.downscale.split("_")
                    if scale[0] == "v":
                        coord, rgb_v = load_s3dis_scene(
                            dir_=coord_file, downscaling=float(scale[1]), rgb=True
                        )
                    else:
                        coord, rgb_v = load_s3dis_scene(
                            dir_=coord_file, random_ds=float(scale[1]), rgb=True
                        )
                else:
                    coord, rgb_v = load_s3dis_scene(
                        dir_=coord_file, downscaling=0, rgb=True
                    )
            else:
                if self.downscale is not None:
                    scale = self.downscale.split("_")
                    if scale[0] == "v":
                        coord = load_s3dis_scene(
                            dir_=coord_file, downscaling=float(scale[1])
                        )
                    else:
                        coord = load_s3dis_scene(
                            dir_=coord_file, random_ds=float(scale[1])
                        )
                else:
                    coord = load_s3dis_scene(dir_=coord_file, downscaling=0)

            if self.downscale is not None:
                if scale[0] == "v":
                    coord[:, 1:] = coord[:, 1:] / float(scale[1])
                else:
                    coord[:, 1:] = coord[:, 1:] / round(
                        pc_median_dist(coord[:, 1:], avg_over=True, box_size=0.25), 3
                    )

            if self.rgb:
                (
                    coords_idx,
                    node_idx,
                    graph_idx,
                    output_idx,
                    cls_idx,
                ) = self.VD.patched_dataset(coord=coord, rgb=rgb_v, mesh=8, random=True)
            else:
                (
                    coords_idx,
                    node_idx,
                    graph_idx,
                    output_idx,
                    cls_idx,
                ) = self.VD.patched_dataset(coord=coord, mesh=8, random=True)

            # save data for faster access later
            if not self.benchmark:
                if self.rgb:
                    self.save_temp(i=i, coord=coord, rgb=rgb_v)
                else:
                    self.save_temp(i=i, coord=coord)

            # Store initial patch size for each data to speed up computation
            self.patch_size[i, 0] = 1
        else:
            # Load pre-process data
            if self.rgb:
                coord, rgb_v = self.load_temp(i, coord=True, rgb=True)
            else:
                coord = self.load_temp(i, coord=True)

            if self.rgb:
                (
                    coords_idx,
                    node_idx,
                    graph_idx,
                    output_idx,
                    cls_idx,
                ) = self.VD.patched_dataset(coord=coord, rgb=rgb_v, mesh=8, random=True)
            else:
                (
                    coords_idx,
                    node_idx,
                    graph_idx,
                    output_idx,
                    cls_idx,
                ) = self.VD.patched_dataset(coord=coord, mesh=8, random=True)

        coords_idx, graph_idx, output_idx, node_idx, cls_idx = self.list_to_tensor(
            coord=coords_idx,
            graph=graph_idx,
            output=output_idx,
            df=node_idx,
            cls=cls_idx,
        )

        if self.benchmark:
            # Output file_name, raw_coord, edge_f, node_f, graph, node_idx, node_class
            return idx, coord, coords_idx, node_idx, graph_idx, output_idx, cls_idx
        # Output edge_f,   node_f, graph,     node_idx,   node_class
        return coords_idx, node_idx, graph_idx, output_idx, cls_idx


def build_dataset(
    dataset_type: Union[str, list],
    dirs: list,
    max_points_per_patch: int,
    downscale=None,
    benchmark=False,
):
    """
    Wrapper for DataLoader

    Function that wraps all data loaders and outputs only one asked for depending
    on a dataset

    Args:
        dataset_type (str): Ask to recognize and process the dataset.
        dirs (list): Ask for a list with the directory given as [train, test].
        max_points_per_patch (int): Max number of points per patch.
        downscale (None, float): Overweight downscale factor
        benchmark (bool): If True construct data for benchmark

    Returns:
        Tuple[torch.DataLoader, torch.DataLoader]: Output DataLoader with
        the specified dataset for training and evaluation.
    """

    if isinstance(dataset_type, list):
        assert len(dataset_type) == 4
        if dataset_type[0] == "simulate":
            if not benchmark:
                dl_train = FilamentSimulateDataset(
                    type_=dataset_type[1],
                    sample_count=int(dataset_type[2]),
                    patch_if=max_points_per_patch,
                    train=True,
                    downscale=downscale,
                )
            dl_test = FilamentSimulateDataset(
                type_=dataset_type[1],
                sample_count=int(dataset_type[3]),
                patch_if=max_points_per_patch,
                train=False,
                downscale=downscale,
            )
    else:
        if dataset_type in ["filament", "MT", "Mem"]:
            if not benchmark:
                dl_train = FilamentDataset(
                    coord_dir=dirs[0],
                    coord_format=(".CorrelationLines.am", ".csv"),
                    patch_if=max_points_per_patch,
                    train=True,
                    downscale=downscale,
                )
            dl_test = FilamentDataset(
                coord_dir=dirs[1],
                coord_format=(".CorrelationLines.am", ".csv"),
                patch_if=max_points_per_patch,
                benchmark=benchmark,
                train=False,
                downscale=downscale,
            )
        elif dataset_type in ["partnet", "PartNet"]:
            if not benchmark:
                dl_train = PartnetDataset(
                    coord_dir=dirs[0],
                    coord_format=".ply",
                    patch_if=max_points_per_patch,
                    train=True,
                    downscale=downscale,
                )
            dl_test = PartnetDataset(
                coord_dir=dirs[1],
                coord_format=".ply",
                patch_if=max_points_per_patch,
                benchmark=benchmark,
                train=False,
                downscale=downscale,
            )
        elif dataset_type in ["scannet", "ScanNetV2"]:
            if not benchmark:
                dl_train = ScannetDataset(
                    coord_dir=dirs[0],
                    coord_format=".ply",
                    patch_if=max_points_per_patch,
                    train=True,
                    downscale=downscale,
                )
            dl_test = ScannetDataset(
                coord_dir=dirs[1],
                coord_format=".ply",
                patch_if=max_points_per_patch,
                benchmark=benchmark,
                train=False,
                downscale=downscale,
            )
        elif dataset_type == "scannet_rgb":
            if not benchmark:
                dl_train = ScannetColorDataset(
                    coord_dir=dirs[0],
                    coord_format=".ply",
                    patch_if=max_points_per_patch,
                    train=True,
                    downscale=downscale,
                )
            dl_test = ScannetColorDataset(
                coord_dir=dirs[1],
                coord_format=".ply",
                patch_if=max_points_per_patch,
                benchmark=benchmark,
                train=False,
                downscale=downscale,
            )
        elif dataset_type in ["stanford", "S3DIS"]:
            if not benchmark:
                dl_train = Stanford3DDataset(
                    coord_dir=dirs[0],
                    coord_format=".txt",
                    patch_if=max_points_per_patch,
                    train=True,
                    downscale=downscale,
                )
            dl_test = Stanford3DDataset(
                coord_dir=dirs[1],
                coord_format=".txt",
                patch_if=max_points_per_patch,
                benchmark=benchmark,
                train=False,
                downscale=downscale,
            )
        elif dataset_type in ["stanford_rgb", "S3DIS_rgb"]:
            if not benchmark:
                dl_train = Stanford3DDataset(
                    coord_dir=dirs[0],
                    coord_format=".txt",
                    rgb=True,
                    patch_if=max_points_per_patch,
                    train=True,
                    downscale=downscale,
                )
            dl_test = Stanford3DDataset(
                coord_dir=dirs[1],
                coord_format=".txt",
                rgb=True,
                patch_if=max_points_per_patch,
                benchmark=benchmark,
                train=False,
                downscale=downscale,
            )
        else:
            # TODO General dataloader
            # if not benchmark:
            #     dl_train = GeneralDataset(coord_dir=dirs[1],
            #                               coord_format=('.ply'),
            #                               downsampling_if=downsampling_if,
            #                               downsampling_rate=downsampling_rate,
            #                               train=True)
            # dl_test = GeneralDataset(coord_dir=dirs[1],
            #                          coord_format=('.ply'),
            #                          downsampling_if=downsampling_if,
            #                          downsampling_rate=downsampling_rate,
            #                          train=False)
            pass

    if not benchmark:
        return DataLoader(dataset=dl_train, shuffle=True, pin_memory=True), DataLoader(
            dataset=dl_test, shuffle=False, pin_memory=True
        )
    return dl_test



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import random
from typing import Optional, Tuple, Union

import numpy as np
import torch
from sklearn.neighbors import NearestNeighbors, KDTree

from tardis_em.utils.errors import TardisError


def pc_median_dist(pc: np.ndarray, avg_over=False, box_size=0.15) -> float:
    """
    Calculate the median distance between KNN points in the point cloud.

    Args:
        pc (np.ndarray): 2D/3D array of the point clouds.
        avg_over (bool): If True, calculate the median position of all points
            and calculate average k-NN for a selected set of points in that area
            (speed up for big point clouds).
        box_size (float): Boundary box size for 'avg_over'.

    Returns:
        float: Median distance between points in given point cloud.
    """
    if isinstance(pc, torch.Tensor):
        pc = pc.cpu().detach().numpy()

    if avg_over:
        # Build BB and offset by 10% from the border
        box_dim = pc.shape[1]

        if box_dim in [2, 3]:
            min_x = np.min(pc[:, 0])
            max_x = np.max(pc[:, 0])
            offset_x = (max_x - min_x) * box_size

            min_y = np.min(pc[:, 1])
            max_y = np.max(pc[:, 1])
            offset_y = (max_y - min_y) * box_size
        else:
            offset_x = 0
            offset_y = 0

        if box_dim == 3:
            min_z = np.min(pc[:, 2])
            max_z = np.max(pc[:, 2])
            offset_z = (max_z - min_z) * box_size
        else:
            offset_z = 0

        x = np.median(pc[:, 0])
        y = np.median(pc[:, 1])

        if box_dim == 3:
            z = np.median(pc[:, 2])
        else:
            z = 0

        voxel = point_in_bb(
            pc,
            min_x=x - offset_x,
            max_x=x + offset_x,
            min_y=y - offset_y,
            max_y=y + offset_y,
            min_z=z - offset_z,
            max_z=z + offset_z,
        )
        pc = pc[voxel]

    # build a NearestNeighbors object for efficient nearest neighbor search
    nn = NearestNeighbors(n_neighbors=2, algorithm="kd_tree").fit(pc)

    if pc.shape[0] < 3:
        return 1.0

    distances, _ = nn.kneighbors(pc)
    distances = distances[:, 1]

    return float(np.mean(distances))


def point_in_bb(
    points: np.ndarray,
    min_x: int,
    max_x: int,
    min_y: int,
    max_y: int,
    min_z: Optional[np.float32] = None,
    max_z: Optional[np.float32] = None,
) -> np.ndarray:
    """
    Compute a bounding_box filter on the given points

    Args:
        points (np.ndarray): (n,3) array.
        min_i, max_i (int): The bounding box limits for each coordinate.
            If some limits are missing, the default values are -infinite for
            the min_i and infinite for the max_i.

    Returns:
        np.ndarray[bool]: The boolean mask indicates wherever a point should be
            kept or not.
    """
    bound_x = np.logical_and(points[:, 0] > min_x, points[:, 0] < max_x)
    bound_y = np.logical_and(points[:, 1] > min_y, points[:, 1] < max_y)

    if points.shape[0] == 3:
        if min_z is not None or max_z is not None:
            bound_z = np.logical_and(points[:, 2] > min_z, points[:, 2] < max_z)
        else:
            bound_z = np.asarray([True for _ in points[:, 2]])

        bb_filter = np.logical_and(np.logical_and(bound_x, bound_y), bound_z)
    else:
        bb_filter = np.logical_and(bound_x, bound_y)

    return bb_filter


class DownSampling:
    """
    Base down sampling wrapper
    """

    def __init__(self, voxel=None, threshold=None, labels=True, KNN=False):
        if voxel is None:
            self.sample = threshold
        else:
            self.sample = voxel

        # If true downs sample with class ids. expect [ID x X x Y x (Z)] [[N, 3] or [N, 4]]
        self.labels = labels
        self.KNN = KNN

    @staticmethod
    def pc_down_sample(
        coord: np.ndarray,
        sampling,
        rgb=None,
    ) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        if rgb is not None:
            return coord, rgb
        return coord

    def __call__(
        self,
        coord: Optional[np.ndarray] = list,
        rgb: Optional[Union[np.ndarray, list]] = None,
    ) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        """
        Compute voxel down sampling for entire point cloud at once or if coord is a list
        compute voxel down sampling for each list index.
        """
        ds_pc = []
        ds_rgb = []

        # Assert correct data structure
        if self.labels:
            if coord.shape[1] not in [3, 4]:
                TardisError(
                    "130",
                    "tardis_em/dist_pytorch/utils/utils.py",
                    f"Expected coordinate array with IDs of shape [[N, 3] or [N, 4]], but {coord.shape} was given!",
                )
        else:
            if coord.shape[1] not in [2, 3]:
                TardisError(
                    "130",
                    "tardis_em/dist_pytorch/utils/utils.py",
                    f"Expected coordinate array without IDs of shape [[N, 2] or [N, 3]], but {coord.shape} was given!",
                )

        """Down-sample each instance from list and combine"""
        if isinstance(coord, list):
            # Assert if RGB are of the same structure as coord
            if rgb is not None and not isinstance(rgb, list):
                TardisError(
                    "130",
                    "tardis_em/dist_pytorch/utils/utils.py",
                    "List of coordinates require list of rbg but array was give!",
                )

            # Down sample
            id = 0
            for idx, i in enumerate(coord):
                coord_df = i
                if self.labels:
                    id = coord_df[0, 0]

                if rgb is not None:
                    rgb_df = rgb[idx]

                    coord_df, rgb_df = self.pc_down_sample(
                        coord=coord_df, rgb=rgb_df, sampling=self.sample
                    )
                    ds_rgb.append(rgb_df)
                else:
                    coord_df = self.pc_down_sample(coord=coord_df, sampling=self.sample)

                ds_pc.append(
                    np.hstack((np.repeat(id, len(coord_df)).reshape(-1, 1), coord_df))
                )
                if not self.labels:
                    id += 1
        else:
            """Down-sample entire point cloud at once"""
            if rgb is not None:
                return self.pc_down_sample(coord=coord, rgb=rgb, sampling=self.sample)
            else:
                return self.pc_down_sample(coord=coord, sampling=self.sample)


class VoxelDownSampling(DownSampling):
    """
    Wrapper for down sampling of the point cloud using voxel grid (Based on Open3d library)
    """

    def __init__(self, **kwargs):
        super(VoxelDownSampling, self).__init__(**kwargs)

    def pc_down_sample(
        self, coord: np.ndarray, sampling: float, rgb: Optional[np.ndarray] = None
    ) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        """
        This function takes a set of 3D points and a voxel size and returns the centroids
        of the voxels in which the points are located.

        Args:
            coord (np.ndarray): A numpy array of shape (N, 3) containing the 3D coordinates
            of the input points.
            rgb (np.ndarray): A numpy array of shape (N, 3) containing RGB values of each point.
            sampling (float): The size of each voxel in each dimension.

        Returns:
            voxel_centers (np.ndarray): A numpy array of shape (M, 3) containing the centroids
            of the voxels in which the points are located where M is the number
            of unique voxels.
        """
        if self.labels:
            coord_label = coord
            coord = coord[:, 1:]

        # Find the grid cell index for each point
        voxel_index = np.floor(coord / sampling).astype(np.int32)

        # Compute the unique set of voxel indices
        unique_voxel_index, inverse_index, voxel_counts = np.unique(
            voxel_index, axis=0, return_inverse=True, return_counts=True
        )

        # Compute the centroids of each voxel
        voxel_centers = np.zeros((len(unique_voxel_index), 3))
        np.add.at(voxel_centers, inverse_index, coord)
        voxel_centers /= voxel_counts[:, np.newaxis]

        # Retrieve ID value for down sampled point cloud
        if self.labels or rgb is not None or self.KNN:
            # Build a KDTree from the voxel_centers
            tree = KDTree(coord)

            # Query the KDTree to find the nearest voxel center for each coord point
            _, nearest_voxel_index = tree.query(voxel_centers)
            nearest_voxel_index = np.concatenate(nearest_voxel_index)

            if self.labels and not self.KNN:
                # Compute the color of the nearest voxel center for each down-sampled point
                voxel_centers = np.hstack(
                    (coord_label[nearest_voxel_index, 0].reshape(-1, 1), voxel_centers)
                )
            elif self.labels and self.KNN:
                # Compute the color of the nearest voxel center for each down-sampled point
                voxel_centers = np.hstack(
                    (
                        coord_label[nearest_voxel_index, 0].reshape(-1, 1),
                        coord[nearest_voxel_index, :],
                    )
                )
            elif not self.labels and self.KNN:
                # Compute the color of the nearest voxel center for each down-sampled point
                voxel_centers = coord[nearest_voxel_index, :]

        if rgb is not None:
            # Compute the color of the nearest voxel center for each down-sampled point
            return voxel_centers, rgb[nearest_voxel_index]
        return voxel_centers


class RandomDownSampling(DownSampling):
    """
    Wrapper for random sampling of the point cloud
    """

    def __init__(self, **kwargs):
        super(RandomDownSampling, self).__init__(**kwargs)

    @staticmethod
    def pc_down_sample(
        coord: np.ndarray, sampling, rgb: Optional[np.ndarray] = None
    ) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
        """
        Random picked point to down sample point cloud.
        Return correctly preserve ID class.

        Args:
            coord: Array of point to downs-sample
            rgb: Extra node feature like e.g. RGB values do sample with coord.
            sampling: Lambda function to calculate down-sampling rate or fixed float ratio.

        Returns:
            random_sample (np.ndarray): Down-sample array of points.
        """
        if isinstance(sampling, int) or isinstance(sampling, float):
            rand_keep = int(len(coord) * sampling)
        else:
            rand_keep = sampling(coord)

        if rand_keep != len(coord):
            rand_keep = random.sample(
                range(len(coord)), rand_keep
            )  # Randomly select coords
        else:
            rand_keep = list(range(len(coord)))  # Keep all

        if rgb is None:
            return coord[rand_keep]
        else:
            return coord[rand_keep], rgb[rand_keep]


def check_model_dict(model_dict: dict) -> dict:
    """
    Check and rebuild model structure dictionary to ensure back-compatibility.

    Args:
        model_dict (dict): Model structure dictionary.

    Returns:
        dict: Standardize model structure dictionary.
    """
    new_dict = {}

    for key, value in model_dict.items():
        if key.endswith("type"):
            new_dict["dist_type"] = value
        if key.endswith("out"):
            new_dict["n_out"] = value
        if key.endswith("node_input"):
            new_dict["node_input"] = value
        if key.endswith("node_dim"):
            new_dict["node_dim"] = value
        if key.endswith("edge_dim"):
            new_dict["edge_dim"] = value
        if key.endswith("layers"):
            new_dict["num_layers"] = value
        if key.endswith("heads"):
            new_dict["num_heads"] = value
        if key.endswith("cls"):
            new_dict["num_cls"] = value
        if key.endswith("sigma"):
            if key.startswith("coord"):
                new_dict["coord_embed_sigma"] = value
            if key.startswith("rgb"):
                new_dict["rgb_embed_sigma"] = value
        if key.endswith("dropout") or key.startswith("dropout"):
            new_dict["dropout_rate"] = value
        if key.endswith("structure"):
            new_dict["structure"] = value
        if key.endswith("knn"):
            new_dict["num_knn"] = value
    if "num_cls" not in new_dict:
        new_dict["num_cls"] = None

    if "rgb_embed_sigma" not in new_dict:
        new_dict["rgb_embed_sigma"] = 1.0

    if "num_knn" not in new_dict:
        new_dict["num_knn"] = None

    return new_dict



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import random
import shutil
from os import listdir, mkdir
from os.path import isdir, join
from shutil import copyfile, copytree, move
from typing import Optional

from tardis_em.utils.errors import TardisError


def move_train_dataset(
    dir_: str, coord_format: tuple, with_img: bool, img_format: Optional[tuple] = None
):
    """
    Detect and copy all date to new train directory.

    Train dataset builder. Detected files of specific format and moved to:
    - dir/train/masks
    - dir/train/imgs [optional]

    Args:
        dir_ (str): Directory where the file should be output.
        coord_format (tuple): Format of the coordinate files.
        with_img (bool): If True, expect corresponding image files.
        img_format (tuple, optional): Allowed format that can be used.
    """
    if coord_format == ".txt":
        area_list = [d for d in listdir(dir_) if isdir(join(dir_, d))]
        area_list = [f for f in area_list if not f.startswith((".", "train", "test"))]

        for i in area_list:
            copytree(join(dir_, i), join(dir_, "train", "masks", i))

    if not len([f for f in listdir(dir_) if f.endswith(coord_format)]) > 0:
        TardisError(
            "121",
            "tardis_em/utils/dataset.py",
            f"No coordinate file found in given dir {dir_}",
        )

    idx_coord = [f for f in listdir(dir_) if f.endswith(coord_format)]

    for i in idx_coord:
        copyfile(src=join(dir_, i), dst=join(dir_, "train", "masks", i))

    """Sort coord with images if included"""
    if with_img:
        if not len([f for f in listdir(dir_) if f.endswith(img_format)]) > 0:
            TardisError(
                "121",
                "tardis_em/utils/dataset.py",
                f"No image file found in given dir {dir_}",
            )

        idx_coord = [f for f in listdir(dir_) if f.endswith(img_format)]

        for i in idx_coord:
            copyfile(src=join(dir_, i), dst=join(dir_, "train", "imgs", i))


def build_test_dataset(dataset_dir: str, dataset_no: int, stanford=False):
    """
    Standard builder for test datasets.

    This module building a test dataset from training subset, by moving random
    files from train to test directory.
    Number of files is specified in %.

    Files are saved in dir/test/imgs and dir/test/masks.

    Args:
        dataset_dir (str): Directory with train test folders.
        dataset_no (int): Number of datasets to iterate throw.
        stanford (bool): Marker for stanford S3DIS dataset
    """
    if "test" not in listdir(dataset_dir) and "train" not in listdir(dataset_dir):
        TardisError(
            "122",
            "tardis_em/utils/dataset.py",
            f"Could not find train or test folder in directory {dataset_dir}",
        )

    if stanford:
        mkdir(join(dataset_dir, "test", "masks", "Area_1"))
        mkdir(join(dataset_dir, "test", "masks", "Area_3"))
        mkdir(join(dataset_dir, "test", "masks", "Area_5"))

        move(
            join(dataset_dir, "train", "masks", "Area_1", "office_1"),
            join(dataset_dir, "test", "masks", "Area_1", "office_1"),
        )
        move(
            join(dataset_dir, "train", "masks", "Area_3", "WC_1"),
            join(dataset_dir, "test", "masks", "Area_3", "WC_1"),
        )
        move(
            join(dataset_dir, "train", "masks", "Area_5", "storage_1"),
            join(dataset_dir, "test", "masks", "Area_5", "storage_1"),
        )

    image_list = listdir(join(dataset_dir, "train", "imgs"))
    image_list = [i for i in image_list if i.endswith((".tif", ".mrc"))]

    if len(image_list) < 1000:
        # Get list of dataset
        images = []
        for i in range(dataset_no):
            df_imgs = [img for img in image_list if img.startswith(f"{i}")]
            images.append(df_imgs)

        # For each image select 20% random patches
        images = [
            i
            for id_, i in enumerate(images)
            if id_ in random.sample(range(0, len(images)), int(len(images) // 5))
        ]

        for i in images:
            list_move = []
            for j in random.sample(range(0, len(i) - 1), 4 if len(i) > 10 else 0):
                list_move.append(i[j])

            for j in list_move:
                shutil.move(
                    join(dataset_dir, "train", "imgs", j),
                    join(dataset_dir, "test", "imgs", j),
                )

                shutil.move(
                    join(dataset_dir, "train", "masks", j[:-4] + "_mask" + j[-4:]),
                    join(dataset_dir, "test", "masks", j[:-4] + "_mask" + j[-4:]),
                )
    else:
        if len(image_list) > 62500:
            images = random.sample(range(0, len(image_list)), 1000)
        else:
            images = random.sample(
                range(0, len(image_list)), int(len(image_list) // 25)
            )

        for i in images:
            j = image_list[i]

            shutil.move(
                join(dataset_dir, "train", "imgs", j),
                join(dataset_dir, "test", "imgs", j),
            )
            shutil.move(
                join(dataset_dir, "train", "masks", j[:-4] + "_mask" + j[-4:]),
                join(dataset_dir, "test", "masks", j[:-4] + "_mask" + j[-4:]),
            )



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional

import torch
import torch.nn as nn

from tardis_em.cnn.model.init_weights import init_weights
from tardis_em.cnn.utils.utils import number_of_features_per_level


class EncoderBlock(nn.Module):
    """
    ENCODER BUILDER

    Single encoder module composed of nn.MaxPool and convolution module.

    Args:
        in_ch (int): Number of input channels.
        out_ch (int): Number of output channels.
        conv_module (conv_module): Single, Double or RCNN convolution block.
        conv_kernel (int): Convolution kernel size.
        max_pool (int): If True nn.MaxPool is applied.
        pool_kernel (int): Kernel size for max pooling.
        dropout (float, optional): Optionals, dropout rate.
        padding (int): Padding size for the convolution.
        components (str): Components that are used for conv. block.
        num_group (int): Num. of groups for the nn.GroupNorm.
            None -> if nn.GroupNorm is not used.
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        conv_module,
        conv_kernel=3,
        max_pool=True,
        dropout: Optional[float] = None,
        pool_kernel=2,
        padding=1,
        components="3gcr",
        num_group=8,
        attn_features=False,
    ):
        super(EncoderBlock, self).__init__()

        self.attn_features = attn_features
        self.dropout = dropout

        """Optionally, add maxpool"""
        if max_pool:
            if "3" in components:
                self.maxpool = nn.MaxPool3d(kernel_size=pool_kernel)
            elif "2" in components:
                self.maxpool = nn.MaxPool2d(kernel_size=pool_kernel)
        else:
            self.maxpool = None

        """Optionally, add dropout layer"""
        if dropout is not None:
            self.dropout_layer = nn.Dropout(p=dropout)

        """Build CNN block"""
        self.conv_module = conv_module(
            in_ch=in_ch,
            out_ch=out_ch,
            block_type="encoder",
            kernel=conv_kernel,
            padding=padding,
            components=components,
            num_group=num_group,
        )

        if self.attn_features:
            self.attn_conv = conv_module(
                in_ch=in_ch + out_ch,
                out_ch=out_ch,
                block_type="encoder",
                kernel=conv_kernel,
                padding=padding,
                components=components,
                num_group=1 if (in_ch + out_ch) % num_group != 0 else num_group,
            )

        """Initialise the blocks"""
        for m in self.children():
            init_weights(m)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward CNN encoder block.

        Args:
            x (torch.Tensor): Image torch before convolution.

        Returns:
            torch.Tensor: Image after convolution.
        """
        if self.maxpool is not None:
            x = self.maxpool(x)

        x_attn = self.conv_module(x)

        if self.attn_features:
            x_attn = self.attn_conv(torch.cat((x, x_attn), dim=1))

        if self.dropout is not None:
            x_attn = self.dropout_layer(x_attn)

        return x_attn


def build_encoder(
    in_ch: int,
    conv_layers: int,
    conv_layer_scaler: int,
    conv_kernel: int or tuple,
    padding: int or tuple,
    num_group: int,
    components: str,
    pool_kernel: int or tuple,
    conv_module,
    dropout: Optional[float] = None,
    attn_features=False,
) -> nn.ModuleList:
    """
    Encoder wrapper for entire CNN model.

    Create encoder block from feature map and convolution modules. Number of
    encoder layers is indicated by number of features.

    Args:
        in_ch (int): Number of input channels.
        conv_layers (int): Number of convolution layers.
        conv_layer_scaler (int): Number of channel by which each CNN block is scaled up.
        conv_module (conv_module): Single, Double or RCNN convolution block.
        conv_kernel (int): Convolution kernel size.
        pool_kernel (int): Kernel size for max pooling.
        dropout (float, optional): Optionals, dropout rate.
        padding (int): Padding size for the convolution.
        components (str): Components that are used for conv. block.
        num_group (int): Num. of groups for the nn.GroupNorm.
            None -> if nn.GroupNorm is not used.
        attn_features (bool):

    Returns:
        nn.ModuleList: Encoder block.
    """
    encoders = []
    feature_map = number_of_features_per_level(
        channel_scaler=conv_layer_scaler, num_levels=conv_layers
    )

    for i, feature in enumerate(feature_map):
        if i == 0:  # first encoder layer skips max pooling
            encoder = EncoderBlock(
                in_ch=in_ch,
                out_ch=feature,
                conv_module=conv_module,
                conv_kernel=conv_kernel,
                dropout=dropout,
                max_pool=False,
                padding=padding,
                components=components,
                num_group=num_group,
                attn_features=attn_features,
            )
        else:
            encoder = EncoderBlock(
                in_ch=feature_map[i - 1],
                out_ch=feature,
                conv_module=conv_module,
                conv_kernel=conv_kernel,
                dropout=dropout,
                pool_kernel=pool_kernel,
                padding=padding,
                components=components,
                num_group=num_group,
                attn_features=attn_features,
            )
        encoders.append(encoder)

    return nn.ModuleList(encoders)



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import glob
from os import listdir, mkdir, rename
from os.path import isdir, join
from shutil import rmtree
from typing import Union

from tardis_em.utils.errors import TardisError


def build_new_dir(dir_: str):
    """
    Standard set-up for creating new directory for storing all data.

    Args:
        dir_ (str): Directory where folder will be build.
    """
    """ Build temp files directory """
    output = join(dir_, "output")
    image_list = glob.glob(dir_ + "/*.tif")
    if not len(image_list) > 0:
        TardisError(
            "12",
            "tardis_em/utils/setup_envir.py",
            "At least one .tif image has to be in the directory!",
        )

    if not isdir(output):
        mkdir(output)
    else:
        print(
            "Output directory already exist! Files moved to new output_old "
            "directory."
        )
        rename(output, join(dir_, "output_old"))
        mkdir(output)


def build_temp_dir(dir_: str):
    """
    Standard set-up for creating new temp dir for cnn prediction.

    Args:
        dir_ (str): Directory where folder will be build.
    """
    if isdir(join(dir_, "temp")):
        if isdir(join(dir_, "temp", "Patches")) or isdir(
            join(dir_, "temp", "Predictions")
        ):
            clean_up(dir_=dir_)

            mkdir(join(dir_, "temp"))
            mkdir(join(dir_, "temp", "Patches"))
            mkdir(join(dir_, "temp", "Predictions"))
        else:
            mkdir(join(dir_, "temp", "Patches"))
            mkdir(join(dir_, "temp", "Predictions"))
    else:
        mkdir(join(dir_, "temp"))
        mkdir(join(dir_, "temp", "Patches"))
        mkdir(join(dir_, "temp", "Predictions"))

    if not isdir(join(dir_, "Predictions")):
        mkdir(join(dir_, "Predictions"))


def clean_up(dir_: str):
    """
    Clean-up all temp files.

    Args:
        dir_ (str): Main directory where temp dir is located.
    """
    if isdir(join(dir_, "temp")):
        if isdir(join(dir_, "temp", "Patches")):
            rmtree(join(dir_, "temp", "Patches"))

        if isdir(join(dir_, "temp", "Predictions")):
            rmtree(join(dir_, "temp", "Predictions"))

        rmtree(join(dir_, "temp"))


def check_dir(
    dir_: str,
    train_img: str,
    train_mask: str,
    test_img: str,
    test_mask: str,
    with_img: bool,
    img_format: Union[tuple, str],
    mask_format: Union[tuple, str, None],
) -> bool:
    """
    Check list used to evaluate if directory containing dataset for CNN.

    Args:
        dir_ (str): Main directory with all files.
        train_img (str): Directory name with images for training.
        train_mask (str): Directory name with mask images for training.
        img_format (tuple, str): Allowed image format.
        test_img (str): Directory name with images for validation.
        test_mask (str): Directory name with mask images for validation.
        mask_format (tuple, str): Allowed mask image format.
        with_img (bool): GraphFormer bool value for training with/without images.

    Returns:
        bool: Bool value indicating detection of the correct structure dataset
    """
    if mask_format is None:
        return True

    dataset_test = False
    if isdir(join(dir_, "train")) and isdir(join(dir_, "test")):
        dataset_test = True

        if with_img:
            # Check if train img and coord exist and have same files
            if isdir(train_img) and isdir(train_mask):
                if len(
                    [f for f in listdir(train_img) if f.endswith(img_format)]
                ) == len([f for f in listdir(train_mask) if f.endswith(mask_format)]):
                    if (
                        len([f for f in listdir(train_img) if f.endswith(img_format)])
                        == 0
                    ):
                        dataset_test = False
                else:
                    dataset_test = False

            # Check if test img and mask exist and have same files
            if isdir(test_img) and isdir(test_mask):
                if len([f for f in listdir(test_img) if f.endswith(img_format)]) == len(
                    [f for f in listdir(test_mask) if f.endswith(mask_format)]
                ):
                    if (
                        len([f for f in listdir(test_img) if f.endswith(img_format)])
                        == 0
                    ):
                        dataset_test = False
                else:
                    dataset_test = False
        else:
            if isdir(train_img) and isdir(train_mask):
                if len([f for f in listdir(train_mask) if f.endswith(mask_format)]) > 0:
                    pass
                else:
                    dataset_test = False
            else:
                dataset_test = False

            if isdir(test_img) and isdir(test_mask):
                if len([f for f in listdir(test_mask) if f.endswith(mask_format)]) > 0:
                    pass
                else:
                    dataset_test = False
            else:
                dataset_test = False
    return dataset_test



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import os
from os import listdir
from os.path import join, splitext
from typing import Tuple

import numpy as np
import torch
from torch.utils.data import Dataset

from tardis_em.cnn.datasets.augmentation import preprocess
from tardis_em.utils.errors import TardisError
from tardis_em.utils.load_data import load_image
from tardis_em.utils.normalization import MinMaxNormalize


class CNNDataset(Dataset):
    """
    DATASET BUILDER FOR IMAGES AND SEMANTIC LABEL MASKS FOR TRAINING

    Args:
        img_dir (str): Source of the 2D/3D .tif file.
        mask_dir (str): Source of the 2D/3D .tif  images masks.
        size (int): Output patch size for image and mask.
        mask_suffix (str): Suffix name for mask images.
        transform (bool): Call for random transformation on img and mask.
        out_channels (int): Number of output channels.
    """

    def __init__(
        self,
        img_dir: str,
        mask_dir: str,
        size=64,
        mask_suffix="_mask",
        transform=True,
        out_channels=1,
    ):
        self.img_dir = img_dir
        self.mask_dir = mask_dir
        self.size = size
        self.mask_suffix = mask_suffix
        self.transform = transform
        self.out_channels = out_channels
        self.minmax = MinMaxNormalize()

        self.ids = [
            splitext(file)[0] for file in listdir(img_dir) if not file.startswith(".")
        ]
        self.format = [
            splitext(file)[1]
            for file in listdir(img_dir)[:5]
            if not file.startswith(".")
        ][1]

    def __len__(self):
        return len(self.ids)

    def __getitem__(self, i: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Select and process dataset for CNN.

        Args:
            i (int): Image ID number.

        Returns:
            torch.Tensor, torch.Tensor: Tensor of processed image and mask.
        """
        # Find next image and corresponding label mask image
        idx = self.ids[i]
        mask_file = os.path.join(self.mask_dir, str(idx) + "_mask" + self.format)
        img_file = os.path.join(self.img_dir, str(idx) + self.format)

        # Load image and corresponding label mask
        img, _ = load_image(img_file)
        mask, _ = load_image(mask_file)

        if mask.dtype != np.uint8:
            TardisError(
                "147",
                "tardis_em/cnn/dataset/dataloader.py",
                f"Mask should be of np.uint8 dtype but is {mask.dtype}!",
            )

        # Process image and mask
        img, mask = preprocess(
            image=img,
            mask=mask,
            size=self.size,
            transformation=self.transform,
            output_dim_mask=self.out_channels,
        )

        if img.dtype != np.float32 and mask.dtype != np.uint8:
            TardisError(
                "147",
                "tardis_em/cnn/dataset/dataloader.py",
                f"Mask {mask.dtype} and image  {img.dtype} has wrong dtype!",
            )
        if not img.min() >= -1 and not img.max() <= 1:
            TardisError(
                "147",
                "tardis_em/cnn/dataset/dataloader.py",
                "Image file is not binary!",
            )

        return torch.from_numpy(img.copy()).type(torch.float32), torch.from_numpy(
            mask.copy()
        ).type(torch.float32)


class PredictionDataset(Dataset):
    """
    DATASET BUILDER FOR IMAGES AND SEMANTIC LABEL MASKS FOR PREDICTION

    Module has turn off all transformations.

    Args:
        img_dir (str): Source of the 2D/3D .tif file.
        out_channels (int): Number of output channels.
    """

    def __init__(self, img_dir: str, out_channels=1):
        self.img_dir = img_dir
        self.out_channels = out_channels

        self.ids = [
            splitext(file)[0] for file in listdir(img_dir) if not file.startswith(".")
        ]

    def __len__(self):
        return len(self.ids)

    def __getitem__(self, i: int):
        """
        Select and process dataset for CNN.

        Args:
            i (int): Image ID number.

        Returns:
            torch.Tensor, str: Tensor of processed image and image file name.
        """
        idx = self.ids[i]
        img_file = join(self.img_dir, str(idx) + ".tif")

        # Load image
        img, _ = load_image(img_file)
        img = img.astype(np.float32)

        # Process image and mask
        img = preprocess(
            image=img,
            size=img.shape,
            transformation=False,
            output_dim_mask=self.out_channels,
        )

        return torch.from_numpy(img).type(torch.float32), idx



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from typing import Union

import torch
from torch import nn

from tardis_em.dist_pytorch.sparse_model.modules import (
    SparsTriangularUpdate,
)
from tardis_em.dist_pytorch.model.modules import GeluFeedForward


class SparseDistStack(nn.Module):
    """
    Module that stacks multiple layers of SparseDistLayer modules.

    This class allows you to define a deep architecture where each layer is a SparseDistLayer.
    The input to each layer is the output from the previous layer,
    and the output from the final layer is the output of the SparseDistStack.
    """

    def __init__(self, pairs_dim: int, num_layers=1, ff_factor=4, knn=8):
        """
        Initializes the SparseDistStack.

        Args:
            pairs_dim (int): The dimensionality of the pairs in each layer.
            num_layers (int): The number of layers in the stack.
            ff_factor (int): The scale factor for the feed forward network in each layer.
            knn (int): The number of nearest neighbors to consider in each layer.
        """
        super().__init__()
        self.layers = nn.ModuleList()
        for _ in range(num_layers):
            self.layers.append(
                SparseDistLayer(pairs_dim=pairs_dim, ff_factor=ff_factor, knn=knn)
            )

    def __len__(self) -> int:
        """
        Returns the number of layers in the stack.
        """
        return len(self.layers)

    def __getitem__(self, i: int):
        """
        Returns the layer at the specified index.

        Args:
            i (int): The index of the layer to return.

        Returns:
            The layer at the specified index.
        """
        return self.layers[i]

    def forward(
        self, edge_features: torch.tensor, indices: list
    ) -> Union[torch.tensor, list]:
        """
        Forward pass for the SparseDistStack.

        Args:
            edge_features (torch.sparse_coo_tensor): A sparse coordinate
                tensor containing the input data.
            indices (list): List of all indices

        Returns:
            torch.sparse_coo_tensor: A sparse coordinate tensor representing
                the output from the final layer in the stack.
        """
        for layer in self.layers:
            edge_features = layer(h_pairs=edge_features, indices=indices)

        return edge_features


class SparseDistLayer(nn.Module):
    """
    The main sparse distance layer in a PyTorch Module.

    This class implements a layer in a deep learning model that operates on sparse
    coordinate tensors. It applies updates to the edge features of the input data,
    and has placeholder methods for future updates to the node features.
    The edge updates include a row update, a column update, and a feed-forward network.
    """

    def __init__(self, pairs_dim: int, ff_factor=4, knn=8):
        """
        Initializes the SparseDistLayer.

        Args:
            pairs_dim (int): The dimensionality of the pairs in the layer.
            ff_factor (int): The scale factor for the feed forward network.
            knn (int): The number of nearest neighbors to consider in each update.
        """
        super().__init__()
        self.pairs_dim = pairs_dim
        self.channel_dim = int(pairs_dim / 4)
        if self.channel_dim <= 0:
            self.channel_dim = 1

        # ToDo Node optional features update

        # ToDO Edge optional MHA update

        # Edge triangular update
        self.row_update = SparsTriangularUpdate(
            input_dim=pairs_dim, channel_dim=self.channel_dim, axis=1, knn=knn
        )
        self.col_update = SparsTriangularUpdate(
            input_dim=pairs_dim, channel_dim=self.channel_dim, axis=0, knn=knn
        )

        # Edge GeLu FFN normalization layer
        self.pair_ffn = GeluFeedForward(
            input_dim=pairs_dim, ff_dim=pairs_dim * ff_factor
        )

    def update_nodes(self) -> torch.tensor:
        """
        Placeholder method for future node update functionality.
        """
        pass

    def update_edges(
        self, h_pairs: torch.tensor, indices: list
    ) -> Union[torch.tensor, list]:
        """
        Updates edge features by applying row and column updates,
        then applying a feed-forward network.

        Args:
            h_pairs (torch.sparse_coo_tensor): A sparse coordinate tensor
                containing the edge features to be updated.
            indices (list): List of indices

        Returns:
            torch.sparse_coo_tensor: A sparse coordinate tensor
                containing the updated edge features.
        """
        # ToDo Convert node features to edge shape

        # Update edge features
        h_pairs = (
            h_pairs
            + self.row_update(x=h_pairs, indices=indices)
            + self.col_update(x=h_pairs, indices=indices)
        )

        return h_pairs + self.pair_ffn(x=h_pairs)

    def forward(
        self, h_pairs: torch.tensor, indices: list
    ) -> Union[torch.tensor, list]:
        """
        Forward pass for the SparseDistLayer.

        Args:
            h_pairs (torch.sparse_coo_tensor): A sparse coordinate tensor containing the input data.
            indices (list): List of indices

        Returns:
            torch.sparse_coo_tensor: A sparse coordinate tensor
                representing the output from the edge updates.
        """
        # ToDo Update node features and convert to edge shape

        # Update edge features
        h_pairs = self.update_edges(h_pairs=h_pairs, indices=indices)

        return h_pairs



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F


class NodeEmbedding(nn.Module):
    """
    NODE FEATURE EMBEDDING

    Args:
        n_in (int): Number of input features.
        n_out (int): Number of output features.
    """

    def __init__(self, n_in: int, n_out: int, sigma=1):
        super().__init__()

        self.linear = None
        if sigma == 0:
            self.linear = nn.Linear(n_in, n_out, bias=False)
        else:
            self.n_in = n_in
            self.sigma = torch.tensor(sigma, dtype=torch.float32)

            w = torch.randn(n_out, n_in)
            b = torch.rand(n_out) * 2 * torch.pi

            self.register_buffer("weight", w)
            self.register_buffer("bias", b)

    def forward(
        self, input_node: Optional[torch.Tensor] = None
    ) -> Optional[torch.Tensor]:
        """
        Forward node feature embedding.

        Input: Batch x Length x Dim
        Output: Batch x Length x Dim

        Args:
            input_node (torch.Tensor): Node features (RGB or image patches).

        Returns:
            torch.Tensor: Embedded features.
        """
        if input_node is None:
            return None

        if self.linear is not None:
            node = self.linear(input_node)
        else:
            node = torch.cos(F.linear(input_node, self.weight / self.sigma, self.bias))

        return (node + 1) / 2


class EdgeEmbedding(nn.Module):
    """
    COORDINATE EMBEDDING INTO GRAPH

    Set of coordinates is used to build distance matrix which is then
    normalized using negative parabolic function.

    Input: Batch x Length x Dim
    Output: Batch x Length x Length x Dim

    Args:
        n_out (int): Number of features to output.
        sigma (int, optional tuple): Sigma value for an exponential function is
            used to normalize distances.
    """

    def __init__(self, n_out: int, sigma: Union[int, float, list]):
        super().__init__()

        self.n_out = n_out
        self.sigma = sigma

        if isinstance(sigma, (list, tuple)):
            if len(sigma) == 2:
                self._range = torch.linspace(sigma[0], sigma[1], n_out)
            else:
                self._range = torch.linspace(sigma[0], sigma[1], int(sigma[2]))
            assert (
                len(self._range) <= n_out
            ), f"Sigma range is out of shape. n_out = {n_out} but sigma range = {len(self._range)}"

            if len(self._range) != self.n_out:
                self.linear = nn.Linear(len(self._range), n_out, bias=False)
            else:
                self.linear = None
        else:
            self.linear = nn.Linear(1, n_out, bias=False)

    def forward(self, input_coord: torch.Tensor) -> torch.Tensor:
        """
        Forward node feature embedding.

        Args:
            input_coord (torch.Tensor): Edge features ([N, 2] or [N, 3]
                coordinates array).

        Returns:
            torch.Tensor: Embedded features.
        """
        g_len = input_coord.shape[-2]
        g_range = range(g_len)

        dist = torch.cdist(input_coord, input_coord)
        if isinstance(self.sigma, (int, float)):
            dist = torch.exp(-(dist**2) / (self.sigma**2 * 2))

            dist[torch.isnan(dist)] = 0.0
            # isnan = torch.isnan(dist)
            # dist = torch.where(isnan, torch.zeros_like(dist), dist)

            # Overwrite diagonal with 1
            dist = dist.unsqueeze(3)
            dist[:, g_range, g_range, :] = 1.0
        else:
            _range_expanded = self._range.view(1, 1, 1, -1).to(dist.device)
            dist = dist.unsqueeze(-1)
            dist = torch.exp(-(dist**2) / (_range_expanded**2 * 2))
            dist[torch.isnan(dist)] = 0.0
            dist[:, g_range, g_range, :] = 1.0

            # dist_range = torch.zeros(
            #     (1, g_len, g_len, len(self._range)), device=dist.device
            # )
            #
            # for id_1, i in enumerate(self._range):
            #     dist_range[..., id_1] = torch.exp(-(dist**2) / (i**2 * 2))
            #
            # dist_range[torch.isnan(dist_range)] = 0.0
            # dist_range[:, g_range, g_range, :] = 1.0

        if self.linear is not None:
            return self.linear(dist)
        else:
            return dist



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from math import pow, sqrt
from typing import Tuple, Union

import numpy as np
from skimage import draw

from tardis_em.cnn.data_processing.interpolation import interpolation
from tardis_em.utils.errors import TardisError


def draw_instances(
    mask_size: Union[list, tuple],
    coordinate: np.ndarray,
    pixel_size: float,
    circle_size=250,
    label=True,
    dtype=None,
) -> np.ndarray:
    """
    Module to build semantic mask from corresponding coordinates

    Args:
        mask_size (tuple): Size of array that will hold created mask.
        coordinate (np.ndarray): Segmented coordinates of a shape [Label x X x Y x (Z)].
        pixel_size (float): Pixel size in Angstrom.
        circle_size (int): Size of a circle the label mask in Angstrom.
        label (bool): If True, expect label point cloud.
        dtype (dtype):

    Returns:
        np.ndarray: Binary mask with drawn all coordinates as lines.
    """
    if label:
        if coordinate.ndim != 2 and coordinate.shape[1] not in [3, 4]:
            TardisError(
                "113",
                "tardis_em/cnn/data_processing.md/draw_mask.py",
                "Coordinates are of not correct shape, expected: "
                f"shape [Label x X x Y x (Z)] but {coordinate.shape} given!",
            )

    if label:
        label_mask = np.zeros(mask_size, dtype=np.uint16 if dtype is None else dtype)
    else:
        label_mask = np.zeros(mask_size, dtype=np.uint8 if dtype is None else dtype)

    if pixel_size == 0:
        pixel_size = 1

    r = (circle_size // 2) // pixel_size

    if coordinate.shape[1] == 3 or not label:  # Draw 2D mask
        mask_shape = "c"
    else:  # Draw 3D mask
        mask_shape = "s"

    # Number of segments in coordinates
    if label:
        segments = np.unique(coordinate[:, 0])

        """Build semantic mask by drawing circle in 2D for each coordinate point"""
        for i in segments:
            # Pick coordinates for each segment
            points = coordinate[np.where(coordinate[:, 0] == i)[0]][:, 1:]

            label = interpolation(points)
            all_cz, all_cy, all_cx = [], [], []

            """Draw label"""
            for j in range(len(label)):
                c = label[j, :]  # Point center

                if len(c) == 3:
                    cz, cy, cx = draw_mask(
                        r=r, c=c, label_mask=label_mask, segment_shape=mask_shape
                    )
                    all_cz.append(cz)
                    all_cy.append(cy)
                    all_cx.append(cx)
                else:
                    cy, cx = draw_mask(
                        r=r, c=c, label_mask=label_mask, segment_shape=mask_shape
                    )
                    all_cy.append(cy)
                    all_cx.append(cx)

            if len(c) == 3:
                all_cz, all_cy, all_cx = (
                    np.concatenate(all_cz),
                    np.concatenate(all_cy),
                    np.concatenate(all_cx),
                )
                label_mask[all_cz, all_cy, all_cx] = i + 1
            else:
                all_cy, all_cx = (
                    np.concatenate(all_cy),
                    np.concatenate(all_cx),
                )
                label_mask[all_cy, all_cx] = i + 1

        return label_mask
    else:
        # Pick coordinates for each point
        all_cz, all_cy, all_cx = [], [], []

        if np.all(coordinate[:, 2] == 0):
            coordinate = coordinate[:, :2]

        for c in coordinate:
            if len(c) == 3:
                cz, cy, cx = draw_mask(
                    r=r, c=c, label_mask=label_mask, segment_shape=mask_shape
                )
                all_cz.append(cz)
                all_cy.append(cy)
                all_cx.append(cx)
            else:
                cy, cx = draw_mask(
                    r=r, c=c, label_mask=label_mask, segment_shape=mask_shape
                )
                all_cy.append(cy)
                all_cx.append(cx)
                all_cz.append(np.repeat(0, len(cx)))

        all_cz, all_cy, all_cx = (
            np.concatenate(all_cz),
            np.concatenate(all_cy),
            np.concatenate(all_cx),
        )

        if coordinate.shape[1] == 2:
            label_mask[all_cy, all_cx] = 1
        else:
            label_mask[all_cz, all_cy, all_cx] = 1

        return np.where(label_mask == 1, 1, 0).astype(np.uint8)


def draw_semantic_membrane(
    mask_size: tuple, coordinate: np.ndarray, pixel_size: float, spline_size=70
) -> np.ndarray:
    """
    Draw semantic membrane

    For each Z pick individual instance and draw a fitted spline of given thickness.

    Args:
        mask_size (tuple): Size of array that will hold created mask.
        coordinate (np.ndarray): Segmented coordinates of a shape [Label x X x Y x (Z)].
        pixel_size (float): Pixel size in Angstrom.
        spline_size (int): Size of a circle the label mask in Angstrom.

    Returns:
        np.ndarray: Binary mask with drawn all coordinates as lines.
    """
    # Ensure ints
    coordinate = coordinate.astype(np.int32)

    # Initiate mask size
    r = round((spline_size / 2) / pixel_size)

    # Initiate mask
    label_mask = np.zeros(mask_size, dtype=np.uint8)

    """Iterate throw each coord"""
    for i in coordinate:
        c = i[1:]

        cz, cy, cx = draw_mask(r=r, c=c, label_mask=label_mask, segment_shape="s")

        label_mask[cz, cy, cx] = 1

    return label_mask


def draw_instances_membrane(
    mask_size: tuple, coordinate: np.ndarray, pixel_size: float, spline_size=70
) -> np.ndarray:
    """
    Draw instances membrane

    For each Z pick individual instance and draw a fitted spline of given thickness.

    Args:
        mask_size (tuple): Size of array that will hold created mask.
        coordinate (np.ndarray): Segmented coordinates of a shape [Label x X x Y x (Z)].
        pixel_size (float): Pixel size in Angstrom.
        spline_size (int): Size of a circle the label mask in Angstrom.

    Returns:
        np.ndarray: Binary mask with drawn all coordinates as lines.
    """
    # Ensure ints
    coordinate = coordinate.astype(np.int32)

    # Initiate mask size
    r = round((spline_size / 2) / pixel_size)

    # Initiate mask
    label_mask = np.zeros(mask_size, dtype=np.uint16)

    """Iterate throw each coord"""
    for i in coordinate:
        c = i[1:]
        id = i[0]

        cz, cy, cx = draw_mask(r=r, c=c, label_mask=label_mask, segment_shape="s")

        label_mask[cz, cy, cx] = id

    return label_mask


def draw_mask(
    r: int, c: np.ndarray, label_mask: np.ndarray, segment_shape: str
) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """
    Module draw_mask to construct sphere shape of a label

    Args:
        r (int): radius of a circle in Angstrom.
        c (np.ndarray): point in 3D indicating center of a circle [X x Y x Z].
        label_mask (np.ndarray): array of a mask on which circle is drawn.
        segment_shape (str): Type of shape to draw. Expect ['s', 'c'].

    Returns:
        np.ndarray: Binary mask.
    """
    if label_mask.ndim not in [2, 3]:
        TardisError(
            "113",
            "tardis_em/cnn/data_processing.md/draw_mask.py"
            f"Unsupported dimensions given {label_mask.ndim} expected 2!",
        )

    x = int(c[0])
    y = int(c[1])
    if len(c) == 3:
        z = int(c[2])
    else:
        z = None

    if segment_shape not in ["s", "c"]:
        TardisError(
            "123",
            "tardis_em/cnn/data_processing.md/draw_mask.py"
            f"Unsupported shape type to draw given {segment_shape} but "
            'expected "c" - circle or "s" - sphere',
        )
    if segment_shape == "s":
        cz, cy, cx = draw_sphere(r=r, c=(z, y, x), shape=label_mask.shape)
        return cz, cy, cx
    else:
        if z is not None:
            cz, cy, cx = draw_circle(r=r, c=(z, y, x), shape=label_mask.shape)
            return cz, cy, cx
        else:
            cy, cx = draw_circle(r=r, c=(y, x), shape=label_mask.shape)
            return cy, cx


def draw_circle(
    r: int, c: tuple, shape: tuple
) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """
    Draw a circle and shift coordinate to c position.

    Args:
        r (int): radius of a circle in Angstrom.
        c (tuple): point in 3D indicating center of a circle [(Z), Y, X].
        shape (tuple): Shape of mask to eliminated ofe-flowed pixel.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: Tuple of array's with zyx coordinates
    """
    r_dim = round(r * 3)
    c_frame = round(r)

    ny, nx = draw.disk((r, r), r, shape=(r_dim, r_dim))
    if len(c) == 3:
        nz = np.repeat(c[0], len(nx))

        c = ((c[1] - c_frame), (c[2] - c_frame))
        z, y, x = nz, ny + c[0], nx + c[1]

        # Remove pixel out of frame
        zyx = np.array((z, y, x)).T
        del_id = []
        for id_, i in enumerate(zyx):
            if i[0] >= shape[0] or i[1] >= shape[1] or i[2] >= shape[2]:
                del_id.append(id_)

        zyx = np.delete(zyx, del_id, 0)

        return zyx[:, 0], zyx[:, 1], zyx[:, 2]
    else:
        c = ((c[0] - c_frame), (c[1] - c_frame))
        y, x = ny + c[0], nx + c[1]

        # Remove pixel out of frame
        yx = np.array((y, x)).T
        del_id = []
        for id_, i in enumerate(yx):
            if i[0] >= shape[0] or i[1] >= shape[1]:
                del_id.append(id_)

        yx = np.delete(yx, del_id, 0)

        return yx[:, 0], yx[:, 1]


def draw_sphere(
    r: int, c: tuple, shape: tuple
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Draw a sphere and shift coordinate to c position.

    Args:
        r (int): radius of a sphere in Angstrom.
        c (tuple): point in 3D indicating center of a sphere [Z, Y, X].
        shape (tuple): Shape of mask to eliminated ofe-flowed pixel.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: Tuple of array's with zyx coordinates
    """
    r_dim = round(r * 2)
    sphere_frame = np.zeros((r_dim, r_dim, r_dim), dtype=np.int8)
    trim = round(r_dim / 6)
    c_frame = round(r)

    z, y, x = sphere_frame.shape
    for z_dim in range(z):
        for y_dim in range(y):
            for x_dim in range(x):
                dist_zyx_to_c = sqrt(
                    pow(abs(z_dim - c_frame), 2)
                    + pow(abs(y_dim - c_frame), 2)
                    + pow(abs(x_dim - c_frame), 2)
                )
                if dist_zyx_to_c > c_frame:
                    sphere_frame[z_dim, y_dim, x_dim] = False
                else:
                    sphere_frame[z_dim, y_dim, x_dim] = True
    sphere_frame[:trim, :] = False  # Trim bottom of the sphere
    sphere_frame[-trim:, :] = False  # Trim top of the sphere
    z, y, x = np.where(sphere_frame)

    c = ((c[0] - c_frame), (c[1] - c_frame), (c[2] - c_frame))
    z, y, x = z + c[0], y + c[1], x + c[2]

    # Remove pixel out of frame
    zyx = np.array((z, y, x)).T
    del_id = []
    for id_, i in enumerate(zyx):
        if i[0] >= shape[0] or i[1] >= shape[1] or i[2] >= shape[2]:
            del_id.append(id_)
        if i[0] < 0:  # Remove negative value
            del_id.append(id_)

    zyx = np.delete(zyx, del_id, 0)

    return zyx[:, 0], zyx[:, 1], zyx[:, 2]



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import codecs
import time
from datetime import datetime
from io import BytesIO
from typing import List, Optional, Union

import numpy as np

from tardis_em.utils.errors import TardisError
from tardis_em.utils.load_data import mrc_mode, mrc_write_header, MRCHeader
from tardis_em.utils.spline_metric import reorder_segments_id
from tardis_em._version import version
import shutil
from io import StringIO
import os

# STL save exception for Arm64 machines
try:
    import pyvista as pv
except ImportError:
    pass


class NumpyToAmira:
    """
    Builder of Amira file from numpy array.
    Support for 3D only! If 2D data, Z dim build with Z=0
    """

    def __init__(self, as_point_cloud=False):
        self.as_point_cloud = as_point_cloud

        self.tardis_header = [
            "# ASCII Spatial Graph \n",
            "# TARDIS - Transformer And Rapid Dimensionless "
            "Instance Segmentation (R) \n",
            f"# tardis_em-pytorch v{version} \r\n",
            f"# MIT License * 2021-{datetime.now().year} * "
            "Robert Kiewisz & Tristan Bepler \n",
        ]

    def check_3d(
        self, coord: Optional[np.ndarray] = List
    ) -> Union[List[np.ndarray], np.ndarray]:
        """
        Check and correct if needed to 3D

        Args:
            coord (np.ndarray, list): Coordinate file to check for 3D.

        Returns:
            Union[np.ndarray, List[np.ndarray]]: The same or converted to 3D coordinates.
        """
        if isinstance(coord, np.ndarray):
            if coord.shape[1] == 3 and not self.as_point_cloud:
                TardisError(
                    "132",
                    "tardis_em/utils/export_data.py",
                    "Numpy 3D array may not have IDs for each point.",
                )

            if coord.shape[1] == 2:
                if self.as_point_cloud:
                    z = np.repeat(0, len(coord)).reshape(-1, 1)
                    coord = np.hstack((coord, z))
                else:
                    TardisError(
                        "132",
                        "tardis_em/utils/export_data.py",
                        "Numpy 2D array may not have IDs for each point.",
                    )

            if coord.shape[1] == 4 and self.as_point_cloud:
                coord = coord[:, 1:]
        else:
            if not isinstance(coord, list) and not isinstance(coord, tuple):
                TardisError(
                    "130",
                    "tardis_em/utils/export_data.py",
                    "Expected list of np.ndarrays!",
                )

            # Add dummy Z dimension
            coord = [
                np.hstack((c, np.zeros((c.shape[0], 1)))) if c.shape[1] == 3 else c
                for c in coord
            ]

            # Fixed ordering
            ordered_coord = []
            last_id = 0
            for c in coord:
                ordered_c = reorder_segments_id(c)  # ID starts from 0
                max_id = len(np.unique(ordered_c[:, 0])) + last_id

                ordered_coord.append(reorder_segments_id(c, [last_id, max_id]))
                last_id = max_id

            return ordered_coord

        if self.as_point_cloud:
            return coord
        return [reorder_segments_id(coord)]

    @staticmethod
    def _build_labels(labels: Optional[tuple] = None) -> list:
        """
        Build label list

        Args:
            labels (tuple, None): List of labels.

        Returns:
            list: Set of labels.
        """
        label = ["LabelGroup"]

        if labels is None:
            return label
        elif isinstance(labels[0], np.ndarray):
            for i in range(len(labels) - 1):
                label.append(f"LabelGroup{i + 2}")
        elif isinstance(labels[0], str):
            label = labels

        return label

    def _build_header(
        self,
        coord: np.ndarray,
        file_dir: str,
        label: Optional[list] = None,
        score: Optional[list] = None,
    ):
        """
        Standard Amira header builder

        Args:
            coord (np.ndarray): 3D coordinate file.
            file_dir (str): Directory where the file should be saved.
            label (int): If not 0, indicate number of labels.
        """
        # Store common data for the header
        if self.as_point_cloud:
            vertex = int(coord.shape[0])
            edge = 0
            point = 0
        else:
            vertex = int(np.max(coord[:, 0]) + 1) * 2  # Number of splines ends
            edge = int(vertex / 2)  # Number of splines
            point = int(coord.shape[0])  # Total number of points

        # Save header
        with codecs.open(file_dir, mode="w", encoding="utf-8") as f:
            for i in self.tardis_header:
                f.write(i)
            f.write("\n")
            f.write(
                f"define VERTEX {vertex} \n"
                f"define EDGE {edge} \n"
                f"define POINT {point} \n"
            )
            f.write("\n")
            f.write("Parameters { \n")
            if not self.as_point_cloud:
                f.write("    SpatialGraphUnitsVertex { \n")
                for i in label:
                    f.write(
                        f"        {i}" + " { \n"
                        "            Unit -1, \n"
                        "            Dimension -1 \n"
                        "        } \n"
                    )
                f.write("    } \n")
                f.write("    SpatialGraphUnitsEdge { \n")
                for i in label:
                    f.write(
                        f"        {i}" + " { \n"
                        "            Unit -1, \n"
                        "            Dimension -1 \n"
                        "        } \n"
                    )
                f.write("    } \n")
                f.write("    SpatialGraphUnitsPoint { \n")
                f.write("    } \n")
                f.write("    Units { \n" '        Coordinates "Å" \n' "    } \n")
                for id_, i in enumerate(label):
                    f.write(
                        f"    {i}" + " { \n"
                        "		Label0" + " { \n"
                        "			Color 1 0.5 0.5, \n"
                        f"          Id {id_ + 1} \n"
                        "     } \n"
                        "        Id 0, \n"
                        "        Color 1 0 0 \n"
                        "    } \n"
                    )
            f.write('	ContentType "HxSpatialGraph" \n' "} \n")
            f.write("\n")
            f.write(
                "VERTEX { float[3] VertexCoordinates } @1 \n"
                "EDGE { int[2] EdgeConnectivity } @2 \n"
                "EDGE { int NumEdgePoints } @3 \n"
                "POINT { float[3] EdgePointCoordinates } @4 \n"
            )

            if not self.as_point_cloud:
                label_id = 5
                for i in label:
                    f.write("VERTEX { int " + f"{i}" + "} " + f"@{label_id} \n")
                    f.write("EDGE { int " + f"{i}" + "} " + f"@{label_id + 1} \n")
                    label_id += 2

            if score is not None:
                for i in range(score[0]):
                    name_ = score[1][i]
                    label_id = label_id + i
                    f.write("EDGE { float " + f"{name_}" + " } " + f"@{label_id} \n")

            f.write("\n")
            f.write("# Data section follows")

    @staticmethod
    def _write_to_amira(data: list, file_dir: str):
        """
        Recursively write all coordinates point

        Args:
            data (list): List of item's to save recursively.
            file_dir (str): Directory where the file should be saved.
        """
        if not file_dir.endswith(".am"):
            TardisError(
                "133",
                "tardis_em/utils/export_data.py",
                f"{file_dir} must be and .am file!",
            )

        with codecs.open(file_dir, mode="a+", encoding="utf-8") as f:
            f.write("\n")

            for i in data:
                f.write(f"{i} \n")

    def export_amira(
        self,
        file_dir: str,
        coords: Union[tuple, list, np.ndarray] = np.ndarray,
        labels: Union[tuple, list, None] = None,
        scores: Optional[list] = None,
    ):
        """
        Save Amira file with all filaments without any labels

        Args:
            file_dir (str): Directory where the file should be saved.
            coords (np.ndarray, tuple): 3D coordinate file.
            labels (tuple, list, None): Labels names.
            scores (list, None): List of confidence scores for each instance.
        """
        coord_list = self.check_3d(coord=coords)

        if not self.as_point_cloud:
            coords = np.concatenate(coord_list)

        if labels is not None:
            if isinstance(labels, str):
                labels = [labels]

            if len(labels) != len(coord_list):
                TardisError(
                    id_="117",
                    py="tardis_em/utils/export_data.py",
                    desc="Number of labels do not mach number of Arrays!",
                )

        # Build Amira header
        if scores is not None:
            score = len(scores[0])
            self._build_header(
                coord=coords,
                file_dir=file_dir,
                label=self._build_labels(labels),
                score=[score, scores[0]],
            )
        else:
            score = False
            self._build_header(
                coord=coords,
                file_dir=file_dir,
                label=self._build_labels(labels),
                score=None,
            )

        # Save only as a point cloud
        if self.as_point_cloud:
            vertex_coord = ["@1"]
            for c in coords[:, 1:]:
                vertex_coord.append(f"{c[0]:.15e} " f"{c[1]:.15e} " f"{c[2]:.15e}")
            self._write_to_amira(data=vertex_coord, file_dir=file_dir)
        else:
            segments_idx = len(np.unique(coords[:, 0]))
            vertex_id_1 = -2
            vertex_id_2 = -1

            vertex_coord = ["@1"]
            vertex_id = ["@2"]
            point_no = ["@3"]
            point_coord = ["@4"]
            for i in range(segments_idx):
                # Collect segment coord and idx
                segment = coords[np.where(coords[:, 0] == i)[0]][:, 1:]

                # Get Coord for vertex #1 and vertex #2
                vertex = np.array((segment[0], segment[-1:][0]), dtype=object)

                # Append vertex #1 (aka Node #1)
                vertex_coord.append(
                    f"{vertex[0][0]:.15e} "
                    f"{vertex[0][1]:.15e} "
                    f"{vertex[0][2]:.15e}"
                )
                # Append vertex #2 (aka Node #2)
                vertex_coord.append(
                    f"{vertex[1][0]:.15e} "
                    f"{vertex[1][1]:.15e} "
                    f"{vertex[1][2]:.15e}"
                )

                # Get Update id number of vertex #1 and #2
                vertex_id_1 += 2
                vertex_id_2 += 2
                vertex_id.append(f"{vertex_id_1} {vertex_id_2}")

                # Get no. of point in edge
                point_no.append(f"{len(segment)}")

                # Get coord of points in edge
                for j in segment:
                    # Append 3D XYZ coord for point
                    point_coord.append(f"{j[0]:.15e} {j[1]:.15e} {j[2]:.15e}")

            self._write_to_amira(data=vertex_coord, file_dir=file_dir)
            self._write_to_amira(data=vertex_id, file_dir=file_dir)
            self._write_to_amira(data=point_no, file_dir=file_dir)
            self._write_to_amira(data=point_coord, file_dir=file_dir)

            # Write down all labels
            label_id = 5
            vertex_id = 1
            edge_id = 1

            start = 0
            total_vertex = len(np.unique(coords[:, 0])) * 2
            total_edge = len(np.unique(coords[:, 0]))
            for i in coord_list:
                vertex_label = [f"@{label_id}"]
                edge_label = [f"@{label_id + 1}"]

                edge = len(np.unique(i[:, 0]))
                vertex = edge * 2
                if start == 0:  # 1 1 1 1 0 0 0 0 0
                    vertex_label.extend(list(np.repeat(vertex_id, vertex)))
                    vertex_label.extend(list(np.repeat(0, total_vertex - vertex)))

                    edge_label.extend(list(np.repeat(edge_id, edge)))
                    edge_label.extend(list(np.repeat(0, total_edge - edge)))
                else:  # 0 0 0 0 1 1 1 1 1 1 0 0 0
                    vertex_label.extend(list(np.repeat(0, start * 2)))
                    vertex_label.extend(list(np.repeat(vertex_id, vertex)))
                    fill_up = total_vertex - start * 2 - vertex
                    if fill_up > 0:
                        vertex_label.extend(list(np.repeat(0, fill_up)))

                    edge_label.extend(list(np.repeat(0, start)))
                    edge_label.extend(list(np.repeat(edge_id, edge)))
                    fill_up = total_edge - start - edge
                    if fill_up > 0:
                        edge_label.extend(list(np.repeat(0, fill_up)))

                label_id += 2
                vertex_id += 1
                edge_id += 1
                start += edge

                self._write_to_amira(data=vertex_label, file_dir=file_dir)
                self._write_to_amira(data=edge_label, file_dir=file_dir)

            if isinstance(score, int) and score:
                for i, s in enumerate(scores[1]):
                    label_id = label_id + i
                    edge_score = [f"@{label_id}"]

                    for j in range(segments_idx):
                        edge_score.append(f"{s[j]:.15e}")

                    self._write_to_amira(data=edge_score, file_dir=file_dir)


def to_mrc(
    data: np.ndarray, pixel_size: float, file_dir: str, org_header: MRCHeader = None
):
    """
    Save MRC image file

    Args:
        data (np.ndarray): Image file.
        pixel_size (float): Image original pixel size.
        file_dir (str): Directory where the file should be saved.
        org_header(MRCHeader): Optional original header
    """
    mode = mrc_mode(mode=data.dtype, amin=data.min())
    time_ = time.asctime()
    if org_header is not None:
        labels = org_header.labels.split(b"\x00")[0]

        if len(labels) > 0:
            nlabl = org_header.nlabl + 1
            label = f"Predicted with TARDIS {version}                        {time_}"
            labels = labels + b"\x00" + bytes(label, "utf-8")
        else:
            nlabl = 1
            label = f"Predicted with TARDIS {version}                        {time_}"
            labels = bytes(label, "utf-8")
    else:
        nlabl = 1
        label = f"Predicted with TARDIS {version}                        {time_}"
        labels = bytes(label, "utf-8")

    len_ = 800 - len(labels)
    labels = labels + b"\x00" * len_

    if data.ndim == 3:
        dim_ = 3
        zlen, ylen, xlen = np.multiply(data.shape, pixel_size)
        mz, my, mx = data.shape
    else:
        dim_ = 2
        ylen, xlen = np.multiply(data.shape, pixel_size)
        zlen = 1
        my, mx = data.shape
        mz = 1

    header = mrc_write_header(
        data.shape[2] if dim_ == 3 else data.shape[1],  # nx
        data.shape[1] if dim_ == 3 else data.shape[0],  # ny
        data.shape[0] if dim_ == 3 else zlen,  # nz
        mode,  # mrc dtype mode
        0,  # nxstart
        0,  # nystart
        0,  # nzstart
        mx,  # mx
        my,  # my
        mz,  # mz
        xlen,  # xlen
        ylen,  # ylen
        zlen,  # zlen
        90.000,  # alpha
        90.000,  # beta
        90.000,  # gamma
        1,  # mapc
        2,  # mapr
        3,  # maps
        data.min(),  # amin
        data.max(),  # amax
        data.mean(),  # amean
        0,  # ispg, space group 0 means images or stack of images
        0,  # next
        0,  # creatid
        0,  # nint
        0,  # nreal
        0,  # imodStamp
        0,  # imodFlags
        0,  # idtype
        0,  # lens
        0,  # nd1
        0,  # nd2
        0,  # vd1
        0,  # vd2
        0,  # tilt_ox
        0,  # tilt_oy
        0,  # tilt_oz
        0,  # tilt_cx
        0,  # tilt_cy
        0,  # tilt_cz
        0,  # xorg
        0,  # yorg
        0,  # zorg
        b"MAP\x00",  # cmap
        b"DD\x00\x00",  # stamp
        data.std(),  # rms
        nlabl,  # nlabels
        labels,  # labels
    )

    with open(file_dir, "wb") as f:
        # write the header
        f.write(header)

        # write data
        f.write(data.tobytes())


def to_am(data: np.ndarray, pixel_size: float, file_dir: str):
    """
    Save image to binary Amira image file.

    Args:
        data (np.ndarray): Image file.
        pixel_size (float): Image original pixel size.
        file_dir (str): Directory where the file should be saved.
    """
    nz, ny, nx = data.shape
    xLen, yLen, zLen = nx * pixel_size, ny * pixel_size, nz * pixel_size

    am = [
        "# AmiraMesh BINARY-LITTLE-ENDIAN 3.0",
        "# TARDIS - Transformer And Rapid Dimensionless Instance Segmentation (R)",
        f"# tardis_em-pytorch v{version}",
        f"# MIT License * 2021-{datetime.now().year} * Robert Kiewisz & Tristan Bepler",
        "",
        "",
        f"define Lattice {nx} {ny} {nz}",
        "",
        "Parameters {",
        "    Units {",
        '       Coordinates "Å"',
        "    }",
        '    DataWindow "0.000000 255.000000",',
        f'    Content "{nx}x{ny}x{nz} byte, uniform coordinates",',
        f"    BoundingBox 0 {xLen} 0 {yLen} 0 {zLen},",
        '    CoordType "uniform"',
        "}",
        "",
        "Lattice { byte Data } @1",
        "",
        "# Data section follows",
        "@1",
    ]

    with codecs.open(file_dir, mode="w", encoding="utf-8") as f:
        for i in am:
            f.write(f"{i} \n")
    with codecs.open(file_dir, mode="ab+") as f:
        bytes_data = data.flatten().tobytes()
        f.write(BytesIO(bytes_data).getbuffer())


def to_stl(data: np.ndarray, file_dir: str):
    """
    Save a point cloud as a PLY file.

    Parameters:
        data (np.ndarray): The name of the PLY file to create.
        file_dir (str): Output file location.
    """

    def change_first_line_of_file(filename, new_first_line):
        fr = open(filename, "r")
        first_line = fr.readline()
        fr.close()
        first_line_len = len(first_line)

        new_first_line_len = len(new_first_line)
        spaces_num = first_line_len - new_first_line_len
        new_first_line = new_first_line + " " * (spaces_num - 1) + "\n"
        fw = StringIO(new_first_line)
        fr = open(filename, "r+")
        shutil.copyfileobj(fw, fr)
        fr.close()
        fw.close()
        return

    def save_multiblock_stl(multiblock, filename):
        names = multiblock.keys()
        oname, ext = os.path.splitext(filename)
        assert ext == ".stl"

        # each stl file saved (output_filenames)
        ofiles = [f"{oname}_{ii}" + ".stl" for ii in range(len(names))]

        for ii, subpart in enumerate(multiblock):
            subpart.save(ofiles[ii], binary=False)
            change_first_line_of_file(
                ofiles[ii], f"solid {names[ii]}"
            )  # basically changes "solid" to "solid <solid_name>"

        # merge files together
        total_stl = ""
        for fn in ofiles:
            f = open(fn)
            total_stl += f.read()
            f.close()

        # writes total stl file
        with open(oname + ".stl", "w") as f:
            f.write(total_stl)

        # deletes previously written stl files
        for fn in ofiles:
            os.remove(fn)

        return

    cloud = pv.MultiBlock()
    for i in np.unique(data[:, 0]):
        cloud.append(
            pv.PolyData(data[np.where(data[:, 0] == i)[0], 1:]).delaunay_2d(alpha=25),
            f"{i}",
        )

    save_multiblock_stl(cloud, file_dir)



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Union

import torch


def get_device(device: str = "0") -> torch.device:
    """
    Return a device that can be used for training or predictions

    Args:
        device (str, int): Device name or ID.

    Returns:
        torch.device: Device type.
    """
    if device == "gpu":  # Load GPU ID 0
        if torch.cuda.is_available():
            device = torch.device("cuda:0")
        else:
            device = torch.device("cpu")
    elif device == "cpu":  # Load CPU
        device = torch.device("cpu")
    elif device_is_str(device):  # Load specific GPU ID
        if torch.cuda.is_available():
            if int(device) == -1:
                device = torch.device("cpu")
            else:
                device = torch.device(f"cuda:{int(device)}")
        else:
            device = torch.device("cpu")
    elif device == "mps":  # Load Apple silicon
        if torch.backends.mps.is_available():
            device = torch.device("cpu")  # So far pytorch don't support CNN on MPS
        else:
            device = torch.device("cpu")
    return device


def device_is_str(device: str = "0") -> bool:
    """
    Check if used device is convertible to int value

    Args:
        device (str, int): Device ID.

    Returns:
        bool: Check for input string/int
    """
    try:
        int(device)
        if isinstance(int(device), int):
            return True
        else:
            return False
    except ValueError:
        return False



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional, Union

import numpy as np
import torch

from tardis_em.dist_pytorch.utils.utils import VoxelDownSampling
from tardis_em.utils.errors import TardisError
from tardis_em.utils.spline_metric import smooth_spline, sort_segment
from tardis_em.utils.visualize_pc import VisualizeFilaments, VisualizePointCloud


class PropGreedyGraphCut:
    """
    PROBABILITY DRIVEN GREEDY GRAPH CUT

    Perform graph cut on predicted point cloud graph representation using in-coming
    and out-coming edges probability.

    Args:
        threshold (float): Edge connection threshold.
        connection (int): Max allowed number of connections per node.
        smooth (bool): If True, smooth splines.
    """

    def __init__(self, threshold=float, connection=2, smooth=False):
        self.threshold = threshold
        if isinstance(connection, int):
            self.connection = connection
        else:
            self.connection = 10000000
        self.smooth = smooth

    @staticmethod
    def _stitch_graph(graph_pred: list, idx: list) -> np.ndarray:
        """
        Stitcher for graph representation

        Args:
            graph_pred (list): Patches of graph predictions
            idx (list): Idx for each node in patches

        Returns:
             np.ndarray: Stitched graph.
        """
        # Build empty graph
        graph = max([max(f) for f in idx]) + 1
        graph = np.zeros((graph, graph), dtype=np.float32)

        for idx_patch, graph_patch in zip(idx, graph_pred):
            for k, _ in enumerate(idx_patch):
                row = graph_patch[k, :]
                row_v = [
                    (
                        row[id_]
                        if graph[i, idx_patch[k]] == 0
                        else np.mean((graph[i, idx_patch[k]], row[id_]))
                    )
                    for id_, i in enumerate(idx_patch)
                ]

                column = graph_patch[:, k]
                column_v = [
                    (
                        row[id_]
                        if graph[i, idx_patch[k]] == 0
                        else np.mean((graph[i, idx_patch[k]], column[id_]))
                    )
                    for id_, i in enumerate(idx_patch)
                ]

                graph[list(idx_patch), idx_patch[k]] = row_v
                graph[idx_patch[k], list(idx_patch)] = column_v

        return graph

    @staticmethod
    def _stitch_coord(coord: list, idx: list) -> np.ndarray:
        """
        Stitcher for coord in patches.

        Args:
            coord (list): Coords in each patch.
            idx (list): Idx for each node in patches.

        Returns:
             np.ndarray: Stitched coordinates.
        """
        # Conversion to Torch
        if isinstance(coord[0], torch.Tensor):
            coord = [c.cpu().detach().numpy() for c in coord]

        # Build empty coord array
        dim = coord[0].shape[1]
        coord_df = max([max(f) for f in idx]) + 1
        coord_df = np.zeros((coord_df, dim), dtype=np.float32)

        for coord_patch, idx_patch in zip(coord, idx):
            for value, id_ in zip(coord_patch, idx_patch):
                coord_df[id_, :] = value

        return coord_df

    @staticmethod
    def _stitch_cls(cls: list, idx: list) -> np.ndarray:
        """
        Stitcher for nodes in patches.

        Args:
            cls (list): Predicted class in each patch.
            idx (list): Idx for each node in patches.

        Returns:
             np.ndarray: Stitched classed id's.
        """
        # Conversion to Torch
        if isinstance(cls[0], torch.Tensor):
            cls = [c.cpu().detach().numpy() for c in cls]

        # Build empty coord array
        cls_df = max([max(f) for f in idx]) + 1
        cls_df = np.zeros(cls_df, dtype=np.float32)

        for cls_patch, idx_patch in zip(cls, idx):
            # cls_patch = [np.where(i == 1)[0][0] for i in cls_patch]
            for value, id_ in zip(cls_patch, idx_patch):
                cls_df[id_] = value

        return cls_df

    def _adjacency_list(
        self,
        graphs: list,
        coord: np.ndarray,
        output_idx: Optional[list] = None,
        threshold=0.5,
    ) -> Optional[list]:
        """
        Builder of adjacency matrix from stitched coord and graph voxels
        The output of the adjacency matrix is list containing:
        [id][coord][interactions][interaction probability]

        Args:
            graphs (list[np.ndarray]): Graph patch output from DIST.
            coord (np.ndarray): Stitched coord output from DIST.
            output_idx (list[list], None): Index lists from DIST.

        Returns:
            list: Adjacency list of all bind graph connections.
        """
        if output_idx is None:
            if len(graphs) == 1:
                output_idx = list(range(len(graphs[0])))
            else:
                return None

        all_prop = [
            [idx, coord_i.tolist(), [], []] for idx, coord_i in enumerate(coord)
        ]
        for g, o in zip(graphs, output_idx):
            top_k_indices = np.argsort(g, axis=1)
            top_k_probs = np.take_along_axis(g, top_k_indices, axis=1)
            o = np.array(o)

            for i in range(top_k_indices.shape[0]):
                indices = o[top_k_indices[i]].tolist()
                probs = top_k_probs[i].tolist()

                if threshold:
                    filtered_indices_probs = [
                        (idx, prob)
                        for idx, prob in zip(indices, probs)
                        if prob >= self.threshold
                    ]
                else:
                    filtered_indices_probs = [
                        (idx, prob) for idx, prob in zip(indices, probs) if prob != 0
                    ]

                indices, probs = (
                    zip(*filtered_indices_probs) if filtered_indices_probs else ([], [])
                )

                # Remove self-connection
                if o[i] in indices:
                    idx = indices.index(o[i])
                    indices = list(indices)
                    probs = list(probs)
                    del indices[idx]
                    del probs[idx]

                all_prop[o[i]][2].extend(indices)
                all_prop[o[i]][3].extend(probs)

        # Merge duplicates
        for p in all_prop:
            if p[2]:
                unique_indices, inv = np.unique(p[2], return_inverse=True)
                p[2] = unique_indices.tolist()
                p[3] = [
                    np.median(np.array(p[3])[inv == i])
                    for i in range(len(unique_indices))
                ]

                # Sort by probability in descending order and apply connection limit
                sorted_indices = np.argsort(p[3])[::-1]
                p[2] = np.array(p[2])[sorted_indices].tolist()
                p[3] = np.array(p[3])[sorted_indices].tolist()

        return all_prop

    def preprocess_connections(self, adj_matrix):
        """
        Preprocess the adjacency matrix to first ensure mutual connections and then
        limit each node to its top 8 connections based on connection probability.
        """
        # Step 1: Identify potential top connections without immediately limiting to top 8
        potential_top_connections = {}
        for idx, (_, _, connections, probabilities) in enumerate(adj_matrix):
            sorted_indices = sorted(
                range(len(probabilities)), key=lambda k: probabilities[k], reverse=True
            )
            potential_top_connections[idx] = {
                connections[i]: probabilities[i] for i in sorted_indices
            }

        # Step 2: Ensure mutual connections
        mutual_connections = {}
        for idx, connections in potential_top_connections.items():
            for conn_idx, prob in connections.items():
                if idx in potential_top_connections.get(conn_idx, {}):
                    mutual_connections.setdefault(idx, {})[conn_idx] = prob
                    mutual_connections.setdefault(conn_idx, {})[idx] = (
                        potential_top_connections[conn_idx][idx]
                    )

        # Step 3: Limit to top 8 mutual connections based on probability
        for idx, connections in mutual_connections.items():
            top_indices = sorted(connections, key=connections.get, reverse=True)[
                : self.connection
            ]
            adj_matrix[idx][2] = top_indices
            adj_matrix[idx][3] = [connections[i] for i in top_indices]

        return adj_matrix

    def _find_segment_matrix_fast(self, adj_matrix: list) -> list:
        """
        Find a segment of connected nodes considering the connection strength and
        limiting each node to a maximum of 8 connections.
        """
        visited = set()
        to_visit = set()

        # Start with the first node that has connections
        for i, (_, _, connections, _) in enumerate(adj_matrix):
            if connections:
                to_visit.add(i)
                break

        while to_visit:
            current = to_visit.pop()
            if current in visited:
                continue

            visited.add(current)
            _, _, connections, _ = adj_matrix[current]

            for neighbor in connections:
                if neighbor not in visited:
                    neighbor_connections = adj_matrix[neighbor][2]
                    # Check for mutual strong connection
                    if (
                        current in neighbor_connections
                        and len(neighbor_connections) <= self.connection
                    ):
                        to_visit.add(neighbor)

        non_empty_nodes = [i for i in visited if adj_matrix[i][2]]
        return non_empty_nodes

    def _find_segment_matrix(self, adj_matrix: list) -> list:
        """
        !Deprecated!
        Iterative search mechanism that search for connected points in the
        adjacency list.

        Args:
            adj_matrix (list): adjacency matrix of graph connections

        Returns:
            list: List of connected nodes.
        """
        idx_df = [0]
        x = 0

        # Find initial point
        while len(idx_df) == 1 and x < len(adj_matrix):
            idx_df = adj_matrix[x][2][: self.connection]
            idx_df.append(x)
            x += 1

        # Skip if no suitable initial point found
        if x > len(adj_matrix):
            return []

        x -= 1
        new = new_df = adj_matrix[x][2][: self.connection]
        visited = set(idx_df)

        # Pick all point associated with the initial point
        # Check 1: Check if node is not already on the list
        # Check 2: Check if new interaction show up secondary interaction
        # Check 3: Check if found idx's were not associated to previous instance
        while len(new_df) != 0:
            # check only new elements
            new_df = []
            for i in new:
                # Pick secondary interaction for i
                reverse_int = adj_matrix[i][2][: self.connection]

                for j in reverse_int:
                    if j not in visited and i in adj_matrix[j][2][: self.connection]:
                        new_df.append(j)
                        visited.add(j)

            new = new_df

        idx_df = list(visited)

        idx_df = [i for i in idx_df if adj_matrix[i][2] != []]
        return idx_df

    @staticmethod
    def _smooth_segments(coord: np.ndarray) -> np.ndarray:
        """
        Smooth spline by fixed factor.

        Args:
            coord (np.ndarray): Splines in array [ID, X, Y, (Z)].

        Returns:
            np.ndarray: Smooth splines.
        """
        splines = []

        # Smooth spline
        for i in np.unique(coord[:, 0]):
            x = coord[coord[:, 0] == int(i), :]

            if len(x) > 4:
                splines.append(smooth_spline(x, 2))
            else:
                splines.append(x)

        return np.concatenate(splines)

    def patch_to_segment(
        self,
        graph: list,
        coord: Union[np.ndarray, list],
        idx: list,
        prune: int,
        sort=True,
        visualize: Optional[str] = None,
    ) -> np.ndarray:
        """
        Point cloud instance segmentation from graph representation

        From each point cloud (patch) first build adjacency matrix.
        Matrix is then used to iteratively search for new segments.
        For each initial node algorithm search for 2 edges with the highest prop.
        with given threshold. For each found edges, algorithm check if found
        node creates an edge with previous node within 2 highest probability.
        If not alg. search for new edge that fulfill the statement.

        Args:
            graph (list): Graph patch output from Dist.
            coord (np.ndarray): Coordinates for each unsorted point idx.
            idx (list): idx of point included in the segment.
            prune (int): Remove splines composed of number of node given by prune.
            sort (bool): If True sort output.
            visualize (str, None): If not None, visualize output with open3D.
        """
        """Check data"""
        if isinstance(graph, np.ndarray):
            graph = [graph]
        elif isinstance(graph, torch.Tensor):
            graph = [graph.cpu().detach().numpy()]
        elif isinstance(graph, list):
            if isinstance(graph[0], torch.Tensor):
                graph = [g.cpu().detach().numpy() for g in graph]

        if isinstance(idx, np.ndarray):
            idx = [idx.astype(int)]
        elif isinstance(idx, torch.Tensor):
            idx = [idx.cpu().detach().numpy().astype(int)]
        elif isinstance(idx, list):
            if isinstance(idx[0], torch.Tensor):
                idx = [i.cpu().detach().numpy().astype(int) for i in idx]

        if not isinstance(coord, np.ndarray):
            try:
                coord = self._stitch_coord(coord, idx)
            except:
                TardisError(
                    "114",
                    "tardis_em/dist/utils/segment_point_cloud.py",
                    "Coord must be an array of all nodes! "
                    f"Expected list of ndarrays but got {type(coord)}",
                )

        """Build Adjacency list from graph representation"""
        adjacency_matrix = self._adjacency_list(
            graphs=graph, coord=coord, output_idx=idx
        )
        adjacency_matrix = self.preprocess_connections(adjacency_matrix)

        coord_segment = []
        stop = False
        segment_id = 0

        while not stop:
            idx = self._find_segment_matrix_fast(adjacency_matrix)

            """Select segment longer then 3 points"""
            if len(idx) >= prune:
                # Sort points in segment
                if sort:
                    segment = sort_segment(coord=coord[idx])
                else:
                    segment = coord[idx]

                if segment.shape[1] == 3:
                    coord_segment.append(
                        np.stack(
                            (
                                np.repeat(segment_id, segment.shape[0]),
                                segment[:, 0],
                                segment[:, 1],
                                segment[:, 2],
                            )
                        ).T
                    )
                elif segment.shape[1] == 2:
                    coord_segment.append(
                        np.stack(
                            (
                                np.repeat(segment_id, segment.shape[0]),
                                segment[:, 0],
                                segment[:, 1],
                                np.zeros((segment.shape[0],)),
                            )
                        ).T
                    )
                segment_id += 1

            # Mask point assigned to the instance
            for id_ in idx:
                (
                    adjacency_matrix[id_][1],
                    adjacency_matrix[id_][2],
                    adjacency_matrix[id_][3],
                ) = (
                    [],
                    [],
                    [],
                )

            if sum([1 for i in adjacency_matrix if sum(i[2]) > 0]) == 0:
                stop = True

        segments = np.vstack(coord_segment)

        if self.smooth:
            voxel_ds = VoxelDownSampling(voxel=5, labels=True, KNN=True)
            segments = voxel_ds(segments)
            segments = segments[segments[:, 0].argsort()]
            df_coord = []

            for i in np.unique(segments[:, 0]):
                id_ = i
                idx = np.where(segments[:, 0] == id_)[0]

                if len(idx) > 3:
                    df_coord.append(
                        np.hstack(
                            (
                                np.repeat(id_, len(idx)).reshape(-1, 1),
                                sort_segment(segments[idx, 1:]),
                            )
                        )
                    )
            segments = np.concatenate(df_coord)
            segments = self._smooth_segments(segments)

        if visualize is not None:
            if visualize not in ["f", "p"]:
                TardisError(
                    "124",
                    "tardis_em/dist/utils/segment_point_cloud.py",
                    'To visualize output use "f" for filament '
                    'or "p" for point cloud!',
                )

            if visualize == "p":
                VisualizePointCloud(segments, True, None, False)
            elif visualize == "f":
                VisualizeFilaments(segments, False, False)

        return segments



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import logging
import platform
import socket
import sys
from os import get_terminal_size, system

from IPython.core.interactiveshell import InteractiveShell
from IPython.display import clear_output

from tardis_em._version import version


def print_progress_bar(value: int, max_v: int):
    """
    Builder for ASCII type progress bar.

    Args:
        value (int): Current value for the progress bar.
        max_v (int): Maximum number of iterations.
    """
    if is_interactive():
        n_bar = 50
    else:
        try:
            n_bar = get_terminal_size()[0] - 50
        except OSError:
            n_bar = 50

    j = value / max_v
    bar = "█" * int(n_bar * j)
    bar = bar + "-" * int(n_bar * (1 - j))

    return f"[{bar:{n_bar}s}] {int(100 * j)}%  [{value} / {max_v}]"


def is_interactive():
    """
    Simple check if command line window is from Jupiter.
    """
    import __main__ as main

    return not hasattr(main, "__file__")


class TardisLogo:
    """
    BUILDER FOR LOG OUTPUT

    The log output is built to fit into the given window (cmd or Jupyter)
    The side Logo is optional and can be removed upon; set the logo to False
    The title and text are optional.

    Example:

    log = Tardis_Logo()

    log(title='Example',

    text_1='Progress bar:',

    text_2=printProgressBar(value=i, max=len(range(10))),

    ....)
    """

    def __init__(self, logo=True):
        self.title = ""

        self.CLEAR = None
        self.WIDTH = None

        if platform.system() in ["Darwin", "Linux"]:
            self.CLEAR = lambda: print("\x1b[2J\x1b[H")
        else:
            self.CLEAR = lambda: system("cls")

        if is_interactive():
            self.CLEAR = lambda: clear_output(wait=True)

        self.FN = "TARDIS-pytorch Copyright Information:"
        self.C = "Copyright (c) 2021 Robert Kiewisz, Tristan Bepler"

        self.logo = logo

    @staticmethod
    def _build_text(max_=80, text="", repeat=False) -> str:
        """
        Format text input to fit the pre-define window size.

        Args:
            max_ (int): Width of the console window. Defaults to 80.
            text (str): Text value to be included in the output.
            repeat (bool): If True repeat text input till the end of max value.

        Returns:
            str: Formatted text string with max number of characters.
        """
        if repeat:
            text = text * max_

        return f'{text[:max_] + " " * (max_ - len(text))}'

    @staticmethod
    def clear_output(wait=True):
        """
        Clear window output other Jupyter of the command line window.

        Args:
            wait (bool):  Wait to clear the output until the new output is
                available to replace it.
        """
        if InteractiveShell.initialized():
            InteractiveShell.instance().display_pub.clear_output(wait)
        else:
            print("\033[2K\r", end="")
            sys.stdout.flush()
            print("\033[2K\r", end="")
            sys.stderr.flush()

    def cell_width(self):
        """
        Ask for current shell window width
        """
        if is_interactive():  # Jupyter
            self.WIDTH = 90
        else:
            try:  # Console
                self.WIDTH = get_terminal_size()[0] - 5
            except OSError:  # Any other
                self.WIDTH = 100

    def __call__(
        self,
        title="",
        text_0=" ",
        text_1="",
        text_2="",
        text_3="",
        text_4="",
        text_5="",
        text_6="",
        text_7="",
        text_8="",
        text_9="",
        text_10="",
        text_11=" ",
    ):
        """
        Builder call function to output nice looking progress bar.

        Args:
            title (str, optional): Any text string.
            text_1 (str, optional): Any text string.
            text_2 (str, optional): Any text string.
            text_3 (str, optional): Any text string.
            text_4 (str, optional): Any text string.
            text_5 (str, optional): Any text string.
            text_6 (str, optional): Any text string.
            text_7 (str, optional): Any text string.
            text_8 (str, optional): Any text string.
            text_9 (str, optional): Any text string.
            text_10 (str, optional): Any text string.

            text_0, text_11 (str, Optional). Any top and bottom text padding.
        Returns:
            print: Print progress bar with all text options into cmd commend line
                window or jupyter notebook.
        """
        if title != "":
            self.title = title

        # Check and update window size
        self.cell_width()

        self.CLEAR()
        if self.logo:
            logo = (
                f'  {self._build_text(self.WIDTH + 1, "=", True)}\n'
                + f' | {self._build_text(self.WIDTH, "TARDIS  " + version + "  " + self.title)}|\n'
                + f' | {self._build_text(self.WIDTH, " ", True)}|\n'
                + f' | {self._build_text(self.WIDTH, "New York Structural Biology Center")}|\n'
                + f' | {self._build_text(self.WIDTH - 13, "Simons Machine Learning Center")} ___         |\n'
                + f' | {self._build_text(self.WIDTH - 21, " ", True)} _______(_@_)_______ |\n'
                + f" | {self._build_text(self.WIDTH - 21, text_0)} |  TARDIS-pytorch | |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_1)} |_________________| |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_2)}  | _____ | _____ |  |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_3)}  | |###| | |###| |  |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_4)}  | |###| | |###| |  |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_5)}  | _____ | _____ |  |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_6)}  | || || | || || |  |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_7)}  | ||_|| | ||_|| |  |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_8)}  | _____ |$_____ |  |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_9)}  | || || | || || |  |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_10)}  | ||_|| | ||_|| |  |\n"
                + f" | {self._build_text(self.WIDTH - 21, text_11)}  | _____ | _____ |  |\n"
                + f' | {self._build_text(self.WIDTH - 21, " ", True)}  | || || | || || |  |\n'
                + f" | {self._build_text(self.WIDTH - 21, self.FN)}  | ||_|| | ||_|| |  |\n"
                + f" | {self._build_text(self.WIDTH - 21, self.C)}  |       |       |  |\n"
                + f' | {self._build_text(self.WIDTH - 21, "MIT License")}  *****************  |\n'
                + f'  {self._build_text(self.WIDTH + 1, "=", True)}\n'
            )
            print(logo)
        else:
            logo = (
                f'  {self._build_text(self.WIDTH + 1, "=", True)}\n'
                + f' | {self._build_text(self.WIDTH, "TARDIS  " + version + "  " + self.title)}|\n'
                + f' | {self._build_text(self.WIDTH, " ", True)}|\n'
                + f' | {self._build_text(self.WIDTH, "New York Structural Biology Center")}|\n'
                + f' | {self._build_text(self.WIDTH, "Simons Machine Learning Center")}|\n'
                + f' | {self._build_text(self.WIDTH, " ", True)}|\n'
                + f" | {self._build_text(self.WIDTH, text_0)}|\n"
                + f" | {self._build_text(self.WIDTH, text_1)}|\n"
                + f" | {self._build_text(self.WIDTH, text_2)}|\n"
                + f" | {self._build_text(self.WIDTH, text_3)}|\n"
                + f" | {self._build_text(self.WIDTH, text_4)}|\n"
                + f" | {self._build_text(self.WIDTH, text_5)}|\n"
                + f" | {self._build_text(self.WIDTH, text_6)}|\n"
                + f" | {self._build_text(self.WIDTH, text_7)}|\n"
                + f" | {self._build_text(self.WIDTH, text_8)}|\n"
                + f" | {self._build_text(self.WIDTH, text_9)}|\n"
                + f" | {self._build_text(self.WIDTH, text_10)}|\n"
                + f" | {self._build_text(self.WIDTH, text_11)}|\n"
                + f' | {self._build_text(self.WIDTH, " ", True)}|\n'
                + f" | {self._build_text(self.WIDTH, self.FN)}|\n"
                + f" | {self._build_text(self.WIDTH, self.C)}|\n"
                + f' | {self._build_text(self.WIDTH, "MIT License")}|\n'
                + f'  {self._build_text(self.WIDTH + 1, "=", True)}\n'
            )
            print(logo)


class ContextFilter(logging.Filter):
    hostname = socket.gethostname()

    def filter(self, record):
        record.hostname = ContextFilter.hostname
        return True



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from tardis_em.cnn.utils.build_cnn import FNet, ResUNet, UNet, UNet3Plus
from tardis_em.utils.errors import TardisError


def build_cnn_network(
    network_type: str, structure: dict, img_size: int, prediction: bool
):
    """
    Wrapper for building CNN model

    Wrapper take CNN parameter and predefined network type (e.g. unet, etc.),
    and build CNN model.

    Args:
        network_type (str): Name of network [unet, resunet, unet3plus, fnet].
        structure (dict): Dictionary with all setting to build CNN.
        img_size (int): Image patch size used for CNN.
        prediction (bool): If true, build CNN in prediction patch.
    """
    if network_type not in ["unet", "resunet", "unet3plus", "fnet", "fnet_attn"]:
        TardisError(
            "141",
            "tardis_em/cnn/cnn.py",
            f"Wrong CNN network name {network_type}",
        )

    if network_type == "unet":
        return UNet(
            in_channels=structure["in_channel"],
            out_channels=structure["out_channel"],
            img_patch_size=img_size,
            dropout=structure["dropout"],
            conv_kernel=structure["conv_kernel"],
            padding=structure["conv_padding"],
            pool_kernel=structure["maxpool_kernel"],
            num_group=structure["num_group"],
            num_conv_layer=structure["num_conv_layers"],
            conv_layer_scaler=structure["conv_scaler"],
            layer_components=structure["layer_components"],
            prediction=prediction,
        )
    elif network_type == "resunet":
        return ResUNet(
            in_channels=structure["in_channel"],
            out_channels=structure["out_channel"],
            img_patch_size=img_size,
            dropout=structure["dropout"],
            num_conv_layer=structure["num_conv_layers"],
            conv_layer_scaler=structure["conv_scaler"],
            conv_kernel=structure["conv_kernel"],
            padding=structure["conv_padding"],
            pool_kernel=structure["maxpool_kernel"],
            num_group=structure["num_group"],
            layer_components=structure["layer_components"],
            prediction=prediction,
        )
    elif network_type == "unet3plus":
        return UNet3Plus(
            in_channels=structure["in_channel"],
            out_channels=structure["out_channel"],
            classifies=structure["classification"],
            dropout=structure["dropout"],
            img_patch_size=img_size,
            conv_kernel=structure["conv_kernel"],
            padding=structure["conv_padding"],
            pool_kernel=structure["maxpool_kernel"],
            num_conv_layer=structure["num_conv_layers"],
            conv_layer_scaler=structure["conv_scaler"],
            layer_components=structure["layer_components"],
            num_group=structure["num_group"],
            prediction=prediction,
        )
    elif network_type == "fnet":
        return FNet(
            in_channels=structure["in_channel"],
            out_channels=structure["out_channel"],
            img_patch_size=img_size,
            dropout=structure["dropout"],
            conv_kernel=structure["conv_kernel"],
            padding=structure["conv_padding"],
            pool_kernel=structure["maxpool_kernel"],
            num_conv_layer=structure["num_conv_layers"],
            conv_layer_scaler=structure["conv_scaler"],
            layer_components=structure["layer_components"],
            num_group=structure["num_group"],
            prediction=prediction,
            attn_features=False,
        )
    elif network_type == "fnet_attn":
        return FNet(
            in_channels=structure["in_channel"],
            out_channels=structure["out_channel"],
            img_patch_size=img_size,
            dropout=structure["dropout"],
            conv_kernel=structure["conv_kernel"],
            padding=structure["conv_padding"],
            pool_kernel=structure["maxpool_kernel"],
            num_conv_layer=structure["num_conv_layers"],
            conv_layer_scaler=structure["conv_scaler"],
            layer_components=structure["layer_components"],
            num_group=structure["num_group"],
            prediction=prediction,
            attn_features=True,
        )
    else:
        return None



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import torch
import torch.nn as nn

from tardis_em.dist_pytorch.sparse_model.embedding import SparseEdgeEmbeddingV4
from tardis_em.dist_pytorch.sparse_model.layers import SparseDistStack


class SparseDIST(nn.Module):
    """
    Sparse Distance Transformer model in a PyTorch Module.

    This class implements a transformer model that can handle sparse data efficiently.
    It first uses a SparseEdgeEmbedding layer to embed the input coordinate tensor into
    a higher-dimensional space. The SparseDistStack then applies several transformer layers
    to this embedded tensor.
    Lastly, a SparseLinear layer decodes the result into the final output.
    The forward pass of the model can optionally apply a sigmoid function
    to the final output if the predict attribute is set to True.
    """

    def __init__(
        self,
        n_out=1,
        edge_dim=128,
        num_layers=6,
        knn=8,
        coord_embed_sigma=1.0,
        predict=False,
        device="cpu",
    ):
        """
        Initializes the SparseDIST.

        Args:
            n_out (int): The number of output features.
            edge_dim (int): The dimensionality of the edge features.
            num_layers (int): The number of transformer layers.
            coord_embed_sigma (int, list): The standard deviation for the edge embedding.
            predict (bool): A boolean value to decide whether to apply
                a sigmoid activation to the final output.
        """
        super(SparseDIST, self).__init__()

        self.n_out = n_out
        self.edge_dim = edge_dim
        self.num_layers = num_layers
        self.knn = knn
        self.edge_sigma = coord_embed_sigma
        self.predict = predict

        self.coord_embed = SparseEdgeEmbeddingV4(
            n_out=self.edge_dim,
            sigma=self.edge_sigma,
            knn=self.knn,
            _device=device,
        )

        self.layers = SparseDistStack(
            pairs_dim=self.edge_dim,
            num_layers=self.num_layers,
            ff_factor=4,
            knn=self.knn,
        )

        self.decoder = nn.Linear(in_features=self.edge_dim, out_features=self.n_out)

    def embed_input(self, coords: torch.tensor) -> torch.tensor:
        """
        Embeds the input coordinates using the coord_embed layer.

        Args:
            coords (torch.sparse_coo_tensor): A sparse coordinate tensor
                containing the input coordinates.

        Returns:
            torch.sparse_coo_tensor: A sparse coordinate tensor containing the embedded coordinates.
        """
        x, idx = self.coord_embed(input_coord=coords)

        return x, idx

    def forward(self, coord: torch.tensor) -> torch.tensor:
        """
        Forward pass for the SparseDIST.

        Args:
            coord (torch.tensor): A sparse coordinate tensor containing the input data.

        Returns:
            torch.sparse_coo_tensor: A sparse coordinate tensor
                representing the output from the model.
        """
        # Embed coord [n, 3] coordinates into spares tensor
        edge, idx = self.embed_input(coords=coord)  # List[Indices, Values, Shape]

        # Encode throughout the transformer layers
        edge = self.layers(edge_features=edge, indices=idx)

        # Decoder
        edge = self.decoder(edge)

        if self.predict:
            edge = torch.sigmoid(edge)

        return edge[1:, :], idx
